"""The main screen for the game"""

import random
import math
import pygame
import os

import serge.actor
import serge.visual
import serge.events
import serge.common
import serge.blocks.utils
import serge.blocks.visualblocks
import serge.blocks.behaviours
import serge.blocks.actors

if serge.common.PYMUNK_OK:
    import pymunk
    import serge.physical

from theme import G, theme
import common
import model
import thingui
import goldui
import waveui
import serge.blocks.layout
import serge.zone
import serge.sound


# Modes
M_BATTLES = 'battles'
M_BOSS = 'boss'
M_TRANSITION = 'transition'
M_DEAD = 'dead'
M_WON = 'won'


class MainScreen(serge.blocks.actors.ScreenActor):
    """The logic for the main screen"""
    
    def __init__(self, options):
        """Initialise the screen"""
        super(MainScreen, self).__init__('item', 'main-screen')
        self.options = options
        self.current_wave = 1
        self.current_battle = 0
        self.mode = M_BATTLES

    def addedToWorld(self, world):
        """Added to the world"""
        super(MainScreen, self).addedToWorld(world)
        #
        background = serge.blocks.utils.addSpriteActorToWorld(world, 'bg', 'bg', 'main-screen-bg', 'background',
            center_position=G('screen-center'))
        #
        self.player = model.Player()
        self.player.auto_click_interval = 0.0#0.1 if self.options.cheat else 0.0
        common.POWER_UPS = model.PowerUpLoader(os.path.join('game', 'powerups.csv'))
        common.PLAYER = self.player
        try:
            self.engine.getWorld('powerup-screen').findActorByName('powerup-screen').resetDisplay()
        except serge.zone.ActorNotFound:
            pass
        #
        #
        self.enemy_loader = model.EnemyLoader(os.path.join('game', 'enemies.csv'))
        self.sounds = [e.class_name.lower() for e in self.enemy_loader.enemies]
        #
        self.player_ui = serge.blocks.utils.addActorToWorld(
            world,
            thingui.PlayerUI('player', 'player', self.player),
            center_position=G('player-posn'),
        )

        self.s1 = self.player.sidekicks['Sidekick-A']
        self.s2 = self.player.sidekicks['Sidekick-B']
        self.s3 = self.player.sidekicks['Sidekick-C']
        #
        self.s1_ui = serge.blocks.utils.addActorToWorld(
            world,
            thingui.SideKickUI('sidekick', 'sidekick-a', self.s1),
            center_position=G('sidekick-a-posn'),
        )
        self.s2_ui = serge.blocks.utils.addActorToWorld(
            world,
            thingui.SideKickUI('sidekick', 'sidekick-b', self.s2),
            center_position=G('sidekick-b-posn'),
        )
        self.s3_ui = serge.blocks.utils.addActorToWorld(
            world,
            thingui.SideKickUI('sidekick', 'sidekick-c', self.s3),
            center_position=G('sidekick-c-posn'),
        )
        self.s1_ui.setAlive(False)
        self.s2_ui.setAlive(False)
        self.s3_ui.setAlive(False)
        #
        self.stage = serge.blocks.layout.HorizontalBar(
            'hbar', 'hbar',
            G('stage-width'),
            G('stage-height'),
        )
        world.addActor(self.stage)
        self.stage.setLayerName('ui')
        self.stage.moveTo(*G('stage-posn'))

        self.gold_ui = serge.blocks.utils.addActorToWorld(
            world,
            goldui.GoldDisplay('gold', 'gold'),
            center_position=G('gold-posn'),
        )

        self.power_up_button = serge.blocks.utils.addSpriteActorToWorld(
            world,
            'button', 'power-up-button',
            'power-up-button',
            'ui', 
            center_position=G('power-up-button-pos')
        )
        self.power_up_button.linkEvent(serge.events.E_LEFT_CLICK, self.clickPowerUp)

        self.next_battle_button = serge.blocks.utils.addSpriteActorToWorld(
            world,
            'button', 'next-battle-button',
            'next-battle-button',
            'ui',
            center_position=G('next-battle-button-pos')
        )
        self.next_battle_button.linkEvent(serge.events.E_LEFT_CLICK, lambda o,a: self.clickOn(self.newBattle))

        self.boss_retreat_button = serge.blocks.utils.addSpriteActorToWorld(
            world,
            'button', 'boss-retreat-button',
            'boss-retreat-button',
            'ui',
            center_position=G('retreat-button-pos')
        )
        self.boss_retreat_button.linkEvent(serge.events.E_LEFT_CLICK, lambda o,a: self.clickOn(self.retreatBoss))
        self.boss_retreat_button.active = False

        self.boss_battle_button = serge.blocks.utils.addSpriteActorToWorld(
            world,
            'button', 'boss-battle-button',
            'boss-battle-button',
            'ui',
            center_position=G('next-wave-button-pos')
        )
        self.boss_battle_button.linkEvent(serge.events.E_LEFT_CLICK, lambda o,a: self.clickOn(self.bossBattle))

        self.next_wave_button = serge.blocks.utils.addSpriteActorToWorld(
            world,
            'button', 'next-wave-button',
            'next-wave-button',
            'ui',
            center_position=G('next-wave-button-pos')
        )
        self.next_wave_button.linkEvent(serge.events.E_LEFT_CLICK, lambda o,a: self.clickOn(self.nextWave))
        self.next_wave_button.active = False

        self.you_died_button = serge.blocks.utils.addSpriteActorToWorld(
            world,
            'button', 'you-died-button',
            'you-died-button',
            'overlay',
            center_position=G('you-died-button-pos')
        )
        self.you_died_button.linkEvent(serge.events.E_LEFT_CLICK, lambda o, a: self.goBack())
        self.you_died_button.active = False

        self.you_won_button = serge.blocks.utils.addSpriteActorToWorld(
            world,
            'button', 'you-won-button',
            'you-won-button',
            'overlay',
            center_position=G('you-won-button-pos')
        )
        self.you_won_button.linkEvent(serge.events.E_LEFT_CLICK, lambda o, a: self.goBack())
        self.you_won_button.active = False

        #
        # Wave indicator
        self.battle_progress = serge.blocks.utils.addActorToWorld(
            world,
            waveui.WavesDisplay('battle', 'battle', 'Battle'),
            center_position=G('battle-posn'),
        )
        self.wave_progress = serge.blocks.utils.addActorToWorld(
            world,
            waveui.WavesDisplay('wave', 'wave', 'Wave'),
            center_position=G('wave-posn'),
        )
        self.battle_button_timer = 0.0
        self.boss_disabler = False

        self.newBattle()

        #
        # Cheating
        if self.options.cheat:
            fps = serge.blocks.utils.addActorToWorld(world,
                serge.blocks.actors.FPSDisplay(G('fps-x'), G('fps-y'), G('fps-colour'), G('fps-size')))

        world.linkEvent(serge.events.E_ACTIVATE_WORLD, self.activateWorld)
        self.activateWorld(None, None)

    def activateWorld(self, obj, arg):
        """Activated"""
        if not serge.sound.Music.isPlayingSong('main-music'):
            serge.sound.Music.play('main-music', loops=-1)
            serge.sound.Music.setVolume(1.0)
            serge.sound.Music.fadeout(5000.0)

    def clickOn(self, method):
        """Click on a button"""
        serge.sound.Sounds.play('click')
        method()

    def clickPowerUp(self, obj, arg):
        """Clicked the power up"""
        self.log.info('Going to power up screen')
        serge.sound.Sounds.play('click')
        #
        # Sync the gold displayed
        screen = self.engine.getWorld('powerup-screen').findActorByName('powerup-screen')
        screen.gold_ui.setGold(self.gold_ui.getGold(), self.gold_ui.target_gold)
        #
        self.engine.setCurrentWorldByName('powerup-screen')

    def updateActor(self, interval, world):
        """Update this actor"""
        super(MainScreen, self).updateActor(interval, world)
        #
        if self.mode != M_DEAD and self.mode != M_WON:
            self.boss_disabler = False
            #
            # Progress the model
            results = self.player.tickClock(interval / 1000.0, self.enemies)
            self.takeActions(results)
            #
            results = self.enemies.tickClock(interval / 1000.0, [self.player])
            self.takeActions(results)
            #
            if not self.enemies.hasSomeAlive() and self.mode != M_TRANSITION:
                if self.mode == M_BATTLES:
                    self.initiateBattleButton()
                else:
                    self.boss_retreat_button.active = False
                    if self.current_wave < 10:
                        self.next_wave_button.active = True
                    else:
                        self.showWon()
            #
            # If we are autoclicking then auto look for gold
            if self.player.auto_click_interval:
                if random.random() < 0.05:
                    items = self.world.findActorsByTag('gold-item')
                    if items:
                        item = random.choice(items)
                        self.collectGold(item, None)
                    items = self.world.findActorsByTag('health-item')
                    if items:
                        item = random.choice(items)
                        self.collectHealth(item, None)
            #
            # Watch for player death
            if not self.player.is_alive:
                self.showDeath()
            #
            # Watch to see if we can afford to buy something
            self.power_up_button.active = common.POWER_UPS.getItemsAvailable(self.gold_ui.getGold())

            #
            # Update the battle button
            if self.next_battle_button.active and self.mode == M_TRANSITION:
                self.battle_button_timer -= interval
                if self.battle_button_timer <= 0:
                    self.newBattle()
                else:
                    index = max(1, int(self.battle_button_timer / 1000.0 + 1))
                    self.next_battle_button.setSpriteName('next-battle-%d-button' % index)
        #
        if self.keyboard.isClicked(pygame.K_ESCAPE):
            self.engine.setCurrentWorldByName('start-screen')
        #
        # Cheats
        if self.options.cheat:
            if self.keyboard.isClicked(pygame.K_b):
                self.newBattle(True)
            if self.keyboard.isClicked(pygame.K_w):
                self.bossBattle(True)
            if self.keyboard.isClicked(pygame.K_d):
                self.player.is_alive = False
            if self.keyboard.isClicked(pygame.K_l):
                self.nextWave(force=True)
            if self.keyboard.isClicked(pygame.K_c):
                self.player.purse += 500
                self.gold_ui.addGold(500)
            if self.keyboard.isClicked(pygame.K_q):
                next_sound = self.sounds.pop(0)
                self.log.info('Playing %s' % next_sound)
                serge.sound.Sounds.play(next_sound)
            if self.keyboard.isClicked(pygame.K_k):
                for enemy in self.enemies:
                    enemy.is_alive = False

    def initiateBattleButton(self):
        """Start the battle button timer"""
        self.next_battle_button.active = True
        if self.current_battle < 10:
            self.mode = M_TRANSITION
            self.battle_button_timer = 3000.0
            self.next_battle_button.setSpriteName('next-battle-3-button')
        else:
            self.next_battle_button.setSpriteName('next-battle-button')

    def updateModelDisplay(self):
        """Update the display as it may have been changed by power ups"""
        for actor in [self.player_ui, self.s1_ui, self.s2_ui, self.s3_ui]:
            actor.updateModelDisplay()

    def takeActions(self, actions):
        """Take all the actions"""
        for action in actions:
            #
            if action:
                self.log.debug(action.getReportString())
            #
            for actor in [self.player_ui, self.s1_ui, self.s2_ui, self.s3_ui] + self.stage.getChildren():
                actor.takeAction(action)

    def newBattle(self, force=False):
        """Start a new battle"""
        if self.next_battle_button.active or force:
            self.mode = M_BATTLES
            self.log.info('New battle clicked')
            #
            self.current_battle = min(10, self.current_battle + 1)
            self.battle_progress.setProgress(self.current_battle)
            self.wave_progress.setProgress(self.current_wave)
            #
            # Visibility of the boss button
            self.boss_battle_button.active = (self.current_battle == 10)
            self.next_battle_button.active = False
            #
            # Remove old actors
            self.stage.removeChildren()
            #
            self.enemies = self.enemy_loader.getEnemyGroup(self.current_wave, G('enemies-in-wave')[self.current_wave - 1])
            #
            for idx, enemy in enumerate(self.enemies):
                self.log.info('Adding enemy %s' % enemy.class_name)
                new_ui = thingui.AnimalUI('animal', enemy.class_name, enemy)
                new_ui.linkEvent(common.E_COLLECT_GOLD, self.collectGold)
                new_ui.linkEvent(common.E_COLLECT_HEALTH, self.collectHealth)
                new_ui.linkEvent(common.E_ENEMY_CLICKED, self.clickOnEnemy)
                self.stage.addActor(new_ui, 'ui')

    def bossBattle(self, force=False):
        """Start the boss battle"""
        if self.boss_battle_button.active or force and not self.boss_disabler:
            self.log.info('Boss battle clicked')
            #
            # Change mode
            self.mode = M_BOSS
            self.boss_retreat_button.active = True
            #
            self.boss_battle_button.active = False
            #
            # Remove old actors
            self.stage.removeChildren()
            #
            all_enemies = self.enemy_loader.getEnemiesInWave(11)
            self.enemies = model.EnemyGroup([all_enemies[self.current_wave - 1]])
            #
            for idx, enemy in enumerate(self.enemies):
                self.log.info('Adding enemy %s' % enemy.class_name)
                new_ui = thingui.AnimalUI('animal', enemy.class_name, enemy)
                new_ui.linkEvent(common.E_COLLECT_GOLD, self.collectGold)
                new_ui.linkEvent(common.E_COLLECT_HEALTH, self.collectHealth)
                new_ui.linkEvent(common.E_ENEMY_CLICKED, self.clickOnEnemy)
                self.stage.addActor(new_ui, 'ui')
            #
            # Reset HP for the player
            self.player.cureAll()

    def retreatBoss(self):
        """Retreat from the boss battle"""
        self.log.info('Retreat clicked')
        self.boss_retreat_button.active = False
        self.mode = M_BATTLES
        self.newBattle(force=True)
        self.boss_disabler = True # Hack because of event issue

    def nextWave(self, force=False):
        """Move to the next wave"""
        if self.next_wave_button.active or force:
            self.log.info('Next wave clicked')
            self.next_wave_button.active = False
            self.current_wave = min(10, self.current_wave + 1)
            self.current_battle = 0
            self.newBattle(True)
            self.mode = M_BATTLES
            self.player.cureAll()

    def collectGold(self, obj, arg):
        """Collecting some gold"""
        self.gold_ui.addGold(obj.getValue('amount'))
        self.world.scheduleActorRemoval(obj)
        serge.sound.Sounds.play('collect-gold')

    def collectHealth(self, obj, arg):
        """Collecting some health"""
        increase = obj.getValue('amount')
        self.player.hp = min(self.player.max_hp, self.player.hp + increase)
        self.world.scheduleActorRemoval(obj)
        serge.sound.Sounds.play('collect-health')

    def clickOnEnemy(self, obj, arg):
        """Clicked on an enemy"""
        results = [self.player.clickOn(obj.model_actor)]
        self.takeActions(results)
        #
        sound = self.player.weapon_name.replace(' ', '-').lower()
        serge.sound.Sounds.play(sound)

    def showDeath(self):
        """Show the death part"""
        self.mode = M_DEAD
        self.next_wave_button.active = False
        self.next_battle_button.active = False
        self.boss_battle_button.active = False
        self.power_up_button.active = False
        self.you_died_button.active = True

    def showWon(self):
        """Show the won part"""
        self.mode = M_WON
        self.next_wave_button.active = False
        self.next_battle_button.active = False
        self.boss_battle_button.active = False
        self.power_up_button.active = False
        self.you_won_button.active = True

    def goBack(self):
        """Go back to the start"""
        serge.sound.Sounds.play('click')
        self.engine.setCurrentWorldByName('start-screen')


def main(options):
    """Create the main logic"""
    #
    # The screen actor
    s = MainScreen(options)
    world = serge.engine.CurrentEngine().getWorld('main-screen')
    world.addActor(s)
    #
    # The behaviour manager
    manager = serge.blocks.behaviours.BehaviourManager('behaviours', 'behaviours')
    world.addActor(manager)
    #
    # Screenshots
    if options.screenshot:
        manager.assignBehaviour(None, 
            serge.blocks.behaviours.SnapshotOnKey(key=pygame.K_s, size=G('screenshot-size')
                , overwrite=False, location='screenshots'), 'screenshots')
    #


