"""Model of the game logic"""

import csv
import copy
import random
import collections


# Damage types
D_PLAYER_CLICK = 'player-click'
D_PLAYER_CRITICAL_CLICK = 'player-critical-click'
D_SIDE_AOF = 'sidekick-aof'
D_SIDE_CLICK = 'sidekick-click'
D_SIDE_CRITICAL_CLICK = 'sidekick-critical-click'
D_SIDE_HEAL = 'sidekick-heal'
D_ENEMY_CLICK = 'enemy-click'
D_ENEMY_CRITICAL_CLICK = 'enemy-critical-click'
D_ENEMY_DOT = 'enemy-dot'


def float_or_empty(s):
    """Return a float from a string or zero if it is empty"""
    if not s:
        return 0.0
    else:
        return float(s)


class StrikeResult(object):
    """Represents the damage done in a click"""

    action = ''
    target = None
    value = 0.0
    damage = 0.0
    actor = None
    damage_type = ''

    def __init__(self, actor, target=None, action='', value=0.0, damage=0.0, damage_type=''):
        """Initialise"""
        self.actor = actor
        self.target = target
        self.action = action
        self.value = value
        self.damage = damage
        self.damage_type = damage_type

    def __str__(self):
        return '%s on %s (%s/%s)' % (self.action, self.target, self.damage, self.value)

    def __nonzero__(self):
        return self.action != ''

    def getReportString(self):
        """Get a result of the report"""
        if self:
            return '%s: %s on %s (hp %s) for %s, value = %s' % (
                str(self.actor),
                self.action,
                str(self.target),
                self.target.hp,
                self.damage,
                self.value,
        )

class LivingThing(object):
    """A living thing in the game"""

    class_name = 'Thing'
    _hp = 0
    ac = 1
    is_alive = True
    click_damage = 0
    critical_click_probability = 0.0
    critical_click_multiplier = 0.0
    damage_per_second = 0.0
    auto_click_interval = 0
    click_aof_damage = 0
    hit_name = 'Hit'
    hit_per_second_name = 'Hit over time'
    damage_taken_over_time = 0.0
    damage_over_time_interval = 2.0
    heal_per_second = 0.0
    value = 0
    weapon_name = ''
    armour_name = ''

    click_damage_type = None
    click_critical_damage_type = None

    timer_randomness = (0.5, 1.5)

    def __init__(self):
        """Initialise the thing"""
        self.attack_timer = 0.0
        self.damage_timer = 0.0
        self.max_hp = self.hp

    @property
    def hp(self):
        return self._hp

    @hp.setter
    def hp(self, value):
        self._hp = value
        self.max_hp = max(self.max_hp, value)

    def cureAll(self):
        """Restore HP and clear any DOT"""
        self.hp = self.max_hp
        self.damage_taken_over_time = 0

    def takeDamage(self, damage_amount):
        """Take damage from"""
        actual_damage = max(0.0, damage_amount - (self.ac - 1))
        self.hp = max(0.0, self.hp - actual_damage)
        if self.hp == 0.0:
            self.is_alive = False
        return actual_damage

    def healDamage(self, damage):
        """Heal some damage"""
        self.hp = min(self.max_hp, self.hp + damage)

    def takeDamagePerSecond(self, damage_amount):
        """Take damage over time"""
        self.damage_taken_over_time += damage_amount

    def clickOn(self, enemy, aof=False):
        """Click on, and damage, an enemy"""
        result = StrikeResult(self)
        #
        # Watch for healing
        if self.heal_per_second:
            enemy.healDamage(self.heal_per_second)
            result.target = enemy
            result.action = 'Heal'
            result.damage = self.heal_per_second
            result.damage_type = D_SIDE_HEAL
        else:
            if random.random() < self.critical_click_probability:
                damage = self.click_damage * self.critical_click_multiplier
                result.action = 'Critical %s' % self.hit_name
                result.damage_type = self.click_critical_damage_type
            else:
                damage = self.click_damage
                result.action = self.hit_name
                result.damage_type = self.click_damage_type
            #
            if aof:
                result.damage_type = D_SIDE_AOF
            #
            # Do immediate damage
            result.damage = enemy.takeDamage(damage)
            result.target = enemy
            #
            # Watch for damage over time
            if self.damage_per_second:
                enemy.takeDamagePerSecond(self.damage_per_second)
                result.action = self.hit_per_second_name
            #
            if not enemy.is_alive:
                result.value = enemy.value
        #
        return result

    def addPowerUp(self, power_up):
        """Add a power up to this thing"""
        self.click_damage += power_up.click_amount
        if power_up.hp_increase:
            self.max_hp += power_up.hp_increase
            self.hp = self.max_hp
        self.ac += power_up.ac_increase
        self.click_aof_damage += power_up.aof_increase
        if power_up.hps_increase:
            self.heal_per_second += power_up.hps_increase
            self.weapon_name = power_up.name
        #
        if power_up.ac_increase:
            self.armour_name = power_up.name
        elif power_up.click_amount or power_up.aof_increase:
            self.weapon_name = power_up.name

    def tickClock(self, interval, enemies):
        """Tick the attack clock"""
        results = []
        #
        # Check for attacking
        self.attack_timer += interval * random.uniform(*self.timer_randomness)
        if self.auto_click_interval and self.attack_timer >= self.auto_click_interval:
            self.attack_timer -= self.auto_click_interval
            #
            # Decide if we do local damage or AOF damage
            if self.click_aof_damage:
                results.extend([self.clickOn(enemy, aof=True) for enemy in self.getTargets(enemies)])
            elif self.heal_per_second:
                results.append(self.clickOn(self.getHealTarget(enemies)))
            else:
                alive = self.getTargets(enemies)
                if alive:
                    results.append(self.clickOn(random.choice(alive)))
                else:
                    results.append(StrikeResult(self))
        else:
            results.append(StrikeResult(self))
        #
        # Check for taking damage
        self.damage_timer += interval
        if self.damage_timer >= self.damage_over_time_interval:
            self.damage_timer -= self.damage_over_time_interval
            if self.damage_taken_over_time:
                damage = self.takeDamage(self.damage_taken_over_time)
                result = StrikeResult(self)
                result.target = self
                result.action = 'DOT'
                result.damage = damage
                result.damage_type = D_ENEMY_DOT
                results.append(result)
        #
        return results

    def getAlive(self, things):
        """Return things that are still alive"""
        return [thing for thing in things if thing.is_alive]

    def getHealTargets(self):
        """Get healing targets"""
        raise NotImplementedError('Method not implemented')

    def getTargets(self, things):
        """Return the valid targets for AOE hits"""
        return self.getAlive(things)


class Player(LivingThing):
    """Represents players"""

    _hp = 100
    click_damage = 2
    critical_click_probability = 0.2
    critical_click_multiplier = 2.0
    purse = 0
    max_purse_size = 100
    ac = 1

    total_spent = 0
    total_power_ups = 0

    click_damage_type = D_PLAYER_CLICK
    click_critical_damage_type = D_PLAYER_CRITICAL_CLICK

    weapon_name = 'Bare hands'
    armour_name = 'Simple Shirt'

    def __init__(self):
        """Initialise the player"""
        super(Player, self).__init__()
        #
        self.sidekicks = collections.OrderedDict()
        self.sidekicks.update([
            ('Sidekick-A', Sidekick('SKA')),
            ('Sidekick-B', Sidekick('SKB')),
            ('Sidekick-C', Sidekick('SKC')),
        ])
        for sidekick in self.sidekicks.values():
            sidekick.is_alive = False

    def __str__(self):
        return 'Player (hp %s, gold %s)' % (self.hp, self.purse)

    def addMoney(self, amount):
        """Add the amount"""
        self.purse += amount
        self.purse = min(max(0, self.purse), self.max_purse_size)

    def addPowerUp(self, power_up):
        """Add a power up"""
        #
        # First purchase it
        power_up.buyItem()
        self.addMoney(-power_up.cost)
        self.total_power_ups += 1
        self.total_spent += power_up.cost
        #
        # Watch for sidekick power ups
        try:
            sidekick = self.sidekicks[power_up.class_name]
        except KeyError:
            super(Player, self).addPowerUp(power_up)
        else:
            sidekick.is_alive = True
            sidekick.addPowerUp(power_up)

    def tickClock(self, interval, enemies):
        """Tick the clock"""
        results = super(Player, self).tickClock(interval, enemies)
        #
        # Now do the sidekicks
        for sidekick in self.sidekicks.values():
            if sidekick.is_alive:
                results.extend(sidekick.tickClock(interval, enemies + [self]))
        #
        return results


class Sidekick(LivingThing):
    """Represents a sidekick"""

    _hp = 1
    auto_click_interval = 2.0
    click_damage_type = D_SIDE_CLICK
    click_critical_damage_type = D_SIDE_CRITICAL_CLICK

    def __init__(self, name='Unknown Sidekick'):
        """Initialise the sidekick"""
        super(Sidekick, self).__init__()
        #
        self.name = name

    def __str__(self):
        return '%s' % self.name

    def getTargets(self, things):
        """Return the valid targets for AOE hits"""
        return [thing for thing in things if not isinstance(thing, Player) and thing.is_alive]

    def getHealTarget(self, things):
        """Return the healing target"""
        for thing in things:
            if isinstance(thing, Player):
                return thing
        else:
            raise ValueError('Did not find the player to heal')


class Enemy(LivingThing):
    """Represents and enemy"""

    _hp = 10
    click_damage_type = D_ENEMY_CLICK
    click_critical_damage_type = D_ENEMY_CRITICAL_CLICK

    @classmethod
    def fromDict(cls, d):
        """Return a new enemy from a dictionary"""
        e = cls()
        e.class_name = d['Class']
        e.real_name = e.class_name
        e.low_name = d['Low']
        e.medium_name = d['Medium']
        e.high_name = d['High']
        e.value = float(d['Value'])
        e.hp = float(d['HP'])
        e.click_damage = float(d['Hit Damage'])
        e.critical_click_multiplier = float(d['Critical Damage multiplier'])
        e.critical_click_probability = float(d['Critical Damage %']) / 100.0
        e.damage_per_second = float(d['Damage per second'])
        e.auto_click_interval = float(d['Click interval'])
        e.hit_name = d['Hit name']
        e.hit_per_second_name = d['Hit per second name']
        e.wave_multipliers = [float_or_empty(d[str(i + 1)]) for i in range(11)]
        #
        return e

    def __str__(self):
        return '%s - %s' % (self.class_name, self.hp)


class PowerUp(object):
    """A power up that can be used"""

    class_name = 'unknown'
    cost = 0.0
    click_amount = 0.0
    aof_increase = 0.0
    hps_increase = 0.0
    hp_increase = 0.0
    ac_increase = 0.0
    is_available = True
    next_item = None
    description = ''
    name = ''

    @classmethod
    def fromDict(cls, d):
        """Return a new power up from a dictionary"""
        p = cls()
        p.name = d['Name']
        p.description = d['Description']
        p.class_name = d['Class']
        p.cost = float(d['Cost'])
        p.click_amount = float_or_empty(d['Click amount'])
        p.aof_increase = float_or_empty(d['AOF Upgrade'])
        p.hps_increase = float_or_empty(d['HPS Upgrade'])
        p.hp_increase = float_or_empty(d['HP Increase'])
        p.ac_increase = float_or_empty(d['AC Increase'])
        #
        return p

    def buyItem(self):
        """Buy this item"""
        self.is_available = False
        if self.next_item:
            self.next_item.is_available = True

    def __str__(self):
        return '%s - %s' % (self.class_name, self.name)


class EnemyLoader(object):
    """Loads enemies from a file"""

    def __init__(self, filename):
        """Initialise the loader and load the enemies"""
        self.filename = filename
        self.load()

    def load(self):
        """Load the enemies"""
        self.enemies = []
        with file(self.filename, 'r') as f:
            reader = csv.DictReader(f)
            for row in reader:
                self.enemies.append(Enemy.fromDict(row))

    def getEnemy(self, index):
        """Return an enemy"""
        return copy.deepcopy(self.enemies[index])

    def getEnemiesInWave(self, wave):
        """Return the enemies in a certain wave"""
        enemies = [e for e in self.enemies if e.wave_multipliers[wave - 1]]
        result = EnemyGroup()
        #
        # Now fix names and hp
        for enemy in enemies:
            total_waves = 0
            for other_wave in range(wave):
                total_waves += enemy.wave_multipliers[other_wave]
            #
            new_enemy = copy.deepcopy(enemy)
            if other_wave == 0:
                new_enemy.real_name = '%s %s' % (new_enemy.low_name, new_enemy.class_name)
            elif other_wave == 1:
                new_enemy.real_name = '%s %s' % (new_enemy.medium_name, new_enemy.class_name)
                new_enemy.hp = int(new_enemy.hp * 1.25)
            elif other_wave == 2:
                new_enemy.real_name = '%s %s' % (new_enemy.high_name, new_enemy.class_name)
                new_enemy.hp = int(new_enemy.hp * 1.5)
            #
            result.append(new_enemy)
        #
        return result

    def getEnemyGroup(self, wave, number):
        """Return a certain number of enemies"""
        group = EnemyGroup()
        enemies = self.getEnemiesInWave(wave)
        for i in range(number):
            group.append(copy.deepcopy(random.choice(enemies)))
        return group


class PowerUpLoader(object):
    """Loads power ups from a file"""

    def __init__(self, filename):
        """Initialise the loader and load the power ups"""
        self.power_ups = []
        self.classes = {}
        with file(filename, 'r') as f:
            reader = csv.DictReader(f)
            for row in reader:
                new_power_up = PowerUp.fromDict(row)
                self.power_ups.append(new_power_up)
                #
                # Add to class
                if new_power_up.class_name in self.classes:
                    existing_items = self.classes[new_power_up.class_name]
                    existing_items[-1].next_item = new_power_up
                    existing_items.append(new_power_up)
                    new_power_up.is_available = False
                else:
                    self.classes[new_power_up.class_name] = PowerUpGroup([new_power_up])

    def getItemsAvailable(self, money):
        """Return all the items available for a certain amount of money"""
        items = []
        for cls in self.classes.values():
            item = cls.getFirstAvailable()
            if item and item.cost <= money:
                items.append(item)
        #
        return items

    def resetPowerUps(self):
        """Reset all power ups"""
        for cls, items in self.classes.iteritems():
            items[0].is_available = True
            for item in items[1:]:
                item.is_available = False


class EnemyGroup(list):
    """A group of enemies"""

    def hasSomeAlive(self):
        """Return True if some are alive"""
        for item in self:
            if item.is_alive:
                return True
        else:
            return False

    def tickClock(self, interval, enemies):
        """Tick the clock of all alive enemies"""
        results = []
        for item in self:
            if item.is_alive:
                results.extend(item.tickClock(interval, enemies))
        return results


class PowerUpGroup(list):
    """A list of power ups in a class"""

    def getFirstAvailable(self):
        """Get the first available item from a group"""
        for item in self:
            if item.is_available:
                return item
        else:
            return None

    def hasItemsAvailable(self):
        """Return True if we have items available"""
        return self.getFirstAvailable() != None