"""Display the UI for a power up"""

import random

import serge.common
import serge.visual
import serge.engine
import serge.events
import serge.blocks.actors
import serge.blocks.layout
import serge.blocks.visualblocks
import serge.actor
from theme import G, theme
import common


class PowerUpUI(serge.actor.MountableActor):
    """Represents a Power Up on the screen"""

    def __init__(self, tag, name, model_actor):
        """Initialise the power up"""
        super(PowerUpUI, self).__init__(tag, name)
        #
        self.model_actor = model_actor
        self.mouse_over_timer = 0.0

    def addedToWorld(self, world):
        """Added to the world"""
        super(PowerUpUI, self).addedToWorld(world)
        #
        self.cost = serge.blocks.actors.FormattedText(
            'cost', 'cost-text',
            format='$%(cost)d',
            colour=G('power-up-cost-text-colour'),
            font_name=G('gold-font-name'),
            font_size=G('power-up-cost-size'),
            justify='left',
            cost=self.model_actor.cost,
        )
        self.cost.setLayerName('ui-front')
        self.mountActor(self.cost, G('power-up-cost-posn'))
        #
        self.pic = serge.actor.Actor('pic', 'pic')
        self.pic.setLayerName('ui-front')
        self.pic.setSpriteName('%s-power-up' % self.model_actor.class_name.replace(' ', '-').lower())
        self.mountActor(self.pic, G('power-up-pic-posn'))
        #
        self.name = serge.blocks.actors.FormattedText(
            'name', 'name-text',
            format='%(pu_name)s',
            colour=G('power-up-text-colour'),
            font_name=G('power-up-font-name'),
            font_size=G('power-up-font-size'),
            justify='left',
            pu_name=self.model_actor.name,
        )
        self.name.setLayerName('ui-front')
        self.mountActor(self.name, G('power-up-name-posn'))
        #
        self.description = serge.blocks.actors.FormattedText(
            'description', 'description-text',
            format='%(pu_description)s',
            colour=G('power-up-text-colour'),
            font_name=G('power-up-font-name'),
            font_size=G('power-up-description-size'),
            justify='left',
            pu_description=self.model_actor.description.replace('$', '\n'),
        )
        self.description.setLayerName('ui-front')
        self.mountActor(self.description, G('power-up-description-posn'))
        #
        self.effects = serge.blocks.actors.FormattedText(
            'effects', 'effects-text',
            format='%(pu_effects)s',
            colour=G('power-up-effects-colour'),
            font_name=G('power-up-font-name'),
            font_size=G('power-up-effects-size'),
            justify='left',
            pu_effects=self.getDescriptionText(),
        )
        self.effects.setLayerName('ui-front')
        self.mountActor(self.effects, G('power-up-effects-posn'))
        #
        self.off_colour = G('power-up-off-colour')[self.model_actor.class_name]
        self.on_colour = G('power-up-on-colour')[self.model_actor.class_name]
        self.rectangle = serge.actor.Actor('rect', 'rect')
        self.rectangle.visual = serge.blocks.visualblocks.Rectangle(
            G('power-up-rect-size'),
            self.off_colour,
        )
        self.rectangle.setLayerName('ui-back')
        self.mountActor(self.rectangle, G('power-up-rect-pos'))
        self.rectangle.linkEvent(serge.events.E_LEFT_CLICK, self.mouseClick)
        #
        self.blanker = serge.actor.Actor('rect', 'rect')
        self.blanker.visual = serge.blocks.visualblocks.Rectangle(
            G('power-up-rect-size'),
            G('power-up-blanker-colour'),
        )
        self.blanker.setLayerName('ui-blanker')
        self.mountActor(self.blanker, G('power-up-rect-pos'))
        #
        self.mouse = world.getEngine().getMouse()
        #

    def getDescriptionText(self):
        """Return the description text"""
        effects = ''
        if self.model_actor.click_amount:
            effects += self.getEffectName('Adds ATK%d' % self.model_actor.click_amount)
        if self.model_actor.aof_increase:
            effects += self.getEffectName('Adds ATK%d AOF' % self.model_actor.aof_increase)
        if self.model_actor.hps_increase:
            effects += self.getEffectName('Adds HP%d heal' % self.model_actor.hps_increase)
        if self.model_actor.hp_increase:
            effects += self.getEffectName('Adds HP%d to max HP' % self.model_actor.hp_increase)
        if self.model_actor.ac_increase:
            effects += self.getEffectName('Adds AC%d to armour' % self.model_actor.ac_increase)
        return effects

    def getEffectName(self, format):
        """Return the effect name"""
        part = format
        if self.model_actor.class_name.startswith('Sidekick'):
            part += ' for sidekick'
        part += '\n'
        return part

    def updateActor(self, interval, world):
        """Update the actor"""
        super(PowerUpUI, self).updateActor(interval, world)
        #
        if self.mouse.getScreenPoint().isInside(self.rectangle):
            self.mouseOver(None, None)
        else:
            self.mouse_over_timer += interval
            if self.mouse_over_timer >= 0 and self.rectangle.visual.colour == self.on_colour:
                self.rectangle.visual.colour = self.off_colour

    def mouseOver(self, obj, arg):
        """Mouse over the button"""
        self.rectangle.visual.colour = self.on_colour
        self.mouse_over_timer = -G('power-up-mouse-time')

    def mouseClick(self, obj, arg):
        """Mouse clicked"""
        if not self.blanker.visible:
            self.log.info('Clicked on power up %s' % self.model_actor)
            self.processEvent((common.E_POWER_UP_BOUGHT, self.model_actor))
            self.updateDisplay()

    def updateDisplay(self):
        """Update the display of the power up"""
        #
        # Update for the new power up
        power_ups = common.POWER_UPS.classes[self.model_actor.class_name]
        next_power_up = power_ups.getFirstAvailable()
        if not next_power_up:
            self.visible = False
        else:
            self.model_actor = next_power_up
            self.name.setValue('pu_name', self.model_actor.name)
            self.description.setValue('pu_description', self.model_actor.description.replace('$', '\n'))
            self.cost.setValue('cost', self.model_actor.cost)
            self.effects.setValue('pu_effects', self.getDescriptionText())

    def setGoldAmount(self, gold):
        """Set the amount of gold we have"""
        self.blanker.visible = gold < self.model_actor.cost