"""Tests for the behaviour trees"""

import unittest
import time

import serge.render
import serge.engine
import serge.actor
import serge.visual
import serge.world
import serge.events
import serge.blocks.utils

from helper import *

from serge.blocks import behaviourtree as bt


class TestBehaviourTrees(unittest.TestCase):
    """Tests for the BehaviourTrees"""

    onStartCalls = []
    onTickCalls = []

    def setUp(self):
        """Set up the tests"""
        self.t = bt.Tree()
        self.c = {
            'True': True,
            'False': False,
        }
        TestBehaviourTrees.onStartCalls = []
        TestBehaviourTrees.onTickCalls = []

    def tearDown(self):
        """Tear down the tests"""

    def testCanAddNodes(self):
        """testCanAddNodes: should be able to add nodes"""
        self.t.addNode(bt.Node('a'))
        self.t.addNode(bt.Node('a'))
        self.t.addNode(bt.Node('a'))
        self.assertEqual(3, self.t.numberOfNodes())

    def testTickCallsAllNodes(self):
        """testTickCallsAllNodes: tick should call tick on all nodes"""
        self.t.addNode(TestNode('a'))
        self.t.addNode(TestNode('b'))
        self.t.addNode(TestNode('c'))
        #
        self.t.onTick(self.c)
        self.assertEqual(['a', 'b', 'c'], [n.name for n in self.onStartCalls])
        self.assertEqual(['a', 'b', 'c'], [n.name for n in self.onTickCalls])

    def testNodeStatus(self):
        """testNodeStatus: check for the correct node status"""
        nodes = [TestNode('a'), TestNode('b')]
        self.t.addNodes(nodes)
        #
        self.assertEqual([bt.Node.S_NONE, bt.Node.S_NONE], [n.status for n in nodes])
        self.t.onTick(self.c)
        self.assertEqual([bt.Node.S_SUCCESS, bt.Node.S_SUCCESS], [n.status for n in nodes])

    def testConditionNode(self):
        """testConditionNode: should be able to create condition nodes"""
        nodes = [bt.Condition('a', lambda c: c['True']), bt.Condition('b', lambda c: c['False'])]
        self.t.addNodes(nodes)
        #
        self.t.onTick(self.c)
        self.assertEqual([bt.Node.S_SUCCESS, bt.Node.S_FAILURE], [n.status for n in nodes])

    def testConditionWithFailure(self):
        """testConditionWithFailure: condition node failing should set appropriate status"""
        nodes = [bt.Condition('a', lambda c: c['NOT_THERE']), bt.Condition('b', lambda c: c['False'])]
        self.t.addNodes(nodes)
        #
        self.t.onTick(self.c)
        self.assertEqual([bt.Node.S_EXCEPTION, bt.Node.S_FAILURE], [n.status for n in nodes])
        self.assertTrue(isinstance(nodes[0].exception, KeyError))

    def testActionNodeSingleCall(self):
        """testActionNodeSingleCall: test for action nodes that complete in one go"""
        n = self.t.addNode(TestAction('a', 1))
        self.t.onTick(self.c)
        self.assertEqual(bt.Node.S_SUCCESS, n.status)
        self.assertEqual(['a'], [n.name for n in self.onTickCalls])

    def testActionNodeMultipleCalls(self):
        """testActionNodeMultipleCalls: test for action nodes taking multiple gos"""
        n = self.t.addNode(TestAction('a', 2))
        self.t.onTick(self.c)
        self.assertEqual(bt.Node.S_RUNNING, n.status)
        self.t.onTick(self.c)
        self.assertEqual(bt.Node.S_SUCCESS, n.status)
        self.assertEqual(['a', 'a'], [n.name for n in self.onTickCalls])

    def testSingleStepAction(self):
        """testSingleStepAction: should be able to do a single step action"""
        n = self.t.addNode(OneStepAction('a'))
        self.t.onTick(self.c)
        self.assertEqual(bt.Node.S_SUCCESS, n.status)
        self.assertEqual([n], self.onTickCalls)

    def testSequenceSuccessTest(self):
        """testSequenceSuccessTest: a simple test of a sequence that succeeds"""
        s = bt.Sequence('s', [
            TestAction('a', 2),
            bt.Condition('c', lambda c: 'True'),
            TestAction('b', 2),
        ])
        self.t.addNode(s)
        self.t.onTick(self.c)
        self.assertEqual(['a'], [n.name for n in self.onTickCalls])
        self.assertEqual(bt.Node.S_RUNNING, s.status)
        self.t.onTick(self.c)
        self.assertEqual(['a', 'a', 'b'], [n.name for n in self.onTickCalls])
        self.assertEqual(bt.Node.S_RUNNING, s.status)
        self.t.onTick(self.c)
        self.assertEqual(['a', 'a', 'b', 'b'], [n.name for n in self.onTickCalls])
        self.assertEqual(bt.Node.S_SUCCESS, s.status)

    def testSequenceFailureTest(self):
        """testSequenceFailureTest: a simple test of a sequence that fails"""
        s = bt.Sequence('s', [
            TestAction('a', 2),
            bt.Condition('c', lambda c: c['False']),
            TestAction('b', 2),
        ])
        self.t.addNode(s)
        self.t.onTick(self.c)
        self.assertEqual(['a'], [n.name for n in self.onTickCalls])
        self.assertEqual(bt.Node.S_RUNNING, s.status)
        self.t.onTick(self.c)
        self.assertEqual(['a', 'a'], [n.name for n in self.onTickCalls])
        self.assertEqual(bt.Node.S_FAILURE, s.status)

    def testEmptySequenceSucceeds(self):
        """testEmptySequenceSucceeds: an empty sequence should immediately succeed"""
        s = bt.Sequence('s', [])
        self.t.addNode(s)
        self.t.onTick(self.c)
        self.assertEqual(bt.Node.S_SUCCESS, s.status)

    def testSelectorAllFail(self):
        """testSelectorAllFail: a selector where all fail"""
        s = bt.Selector('s', [
            bt.Condition('c', lambda c: c['False']),
            TestAction('a', 2, bt.Node.S_FAILURE),
            TestAction('b', 2, bt.Node.S_FAILURE),
        ])
        self.t.addNode(s)
        self.t.onTick(self.c)
        self.assertEqual(bt.Node.S_RUNNING, s.status)
        self.assertEqual(['a'], [n.name for n in self.onTickCalls])
        self.t.onTick(self.c)
        self.assertEqual(bt.Node.S_RUNNING, s.status)
        self.assertEqual(['a', 'a', 'b'], [n.name for n in self.onTickCalls])
        self.t.onTick(self.c)
        self.assertEqual(bt.Node.S_FAILURE, s.status)
        self.assertEqual(['a', 'a', 'b', 'b'], [n.name for n in self.onTickCalls])

    def testSelectorOneSucceeds(self):
        """testSelectorOneSucceeds: a selector where one item succeeds"""
        s = bt.Selector('s', [
            bt.Condition('c', lambda c: c['False']),
            TestAction('a', 2),
            TestAction('b', 2),
        ])
        self.t.addNode(s)
        self.t.onTick(self.c)
        self.assertEqual(bt.Node.S_RUNNING, s.status)
        self.t.onTick(self.c)
        self.assertEqual(bt.Node.S_SUCCESS, s.status)
        self.assertEqual(['a', 'a'], [n.name for n in self.onTickCalls])

    def testEmptySelector(self):
        """testEmptySelector: a selector with no nodes"""
        s = bt.Selector('s', [])
        self.t.addNode(s)
        self.t.onTick(self.c)
        self.assertEqual(bt.Node.S_FAILURE, s.status)

    def testParallelAllFail(self):
        """testParallelAllFail: a parallel node where all children fail"""
        s = bt.Parallel('p', [
            bt.Condition('c', lambda c: c['False']),
            bt.Condition('c', lambda c: c['False']),
            bt.Condition('c', lambda c: c['False']),
        ])
        self.t.addNode(s)
        self.t.onTick(self.c)
        self.assertEqual(bt.Node.S_FAILURE, s.status)

    def testParallelOneLongRunning(self):
        """testParallelOneLongRunning: a parallel node where one is long running"""
        s = bt.Parallel('p', [
            bt.Condition('c', lambda c: c['False']),
            TestAction('a', 2, bt.Node.S_FAILURE),
            bt.Condition('c', lambda c: c['False']),
        ])
        self.t.addNode(s)
        self.t.onTick(self.c)
        self.assertEqual(bt.Node.S_RUNNING, s.status)
        self.assertEqual(['a'], [n.name for n in self.onTickCalls])
        self.t.onTick(self.c)
        self.assertEqual(bt.Node.S_FAILURE, s.status)
        self.assertEqual(['a', 'a'], [n.name for n in self.onTickCalls])

    def testParallelSuccessBeatsFailure(self):
        """testParallelSuccessBeatsFailure: a parallel node where success wins over failure"""
        s = bt.Parallel('s', [
            bt.Condition('c', lambda c: c['False']),
            TestAction('a', 2, bt.Node.S_FAILURE),
            TestAction('b', 2, bt.Node.S_SUCCESS),
            bt.Condition('c', lambda c: c['False']),
        ])
        self.t.addNode(s)
        self.t.onTick(self.c)
        self.assertEqual(bt.Node.S_RUNNING, s.status)
        self.assertEqual(['a', 'b'], [n.name for n in self.onTickCalls])
        self.t.onTick(self.c)
        self.assertEqual(bt.Node.S_SUCCESS, s.status)
        self.assertEqual(['a', 'b', 'a', 'b'], [n.name for n in self.onTickCalls])

    def testParallelEmpty(self):
        """testParallelEmpty: an empty parallel node"""
        n = self.t.addNode(bt.Parallel('p', []))
        self.t.onTick(self.c)
        self.assertEqual(bt.Node.S_FAILURE, n.status)

    def testLoopWithSuccess(self):
        """testLoopWithSuccess: loop node with success"""
        s = self.t.addNode(
            bt.Loop('l',
                bt.Sequence('s', [
                    TestAction('a', 2),
                    TestAction('b', 2),
                ])
            )
        )
        self.t.onTick(self.c)
        self.t.onTick(self.c)
        self.t.onTick(self.c)
        self.assertEqual(bt.Node.S_RUNNING, s.status)   # still running
        self.assertEqual(['a', 'a', 'b', 'b'], [n.name for n in self.onTickCalls])
        self.assertEqual(['a', 'b'], [n.name for n in self.onStartCalls])
        #
        # Now should loop
        self.t.onTick(self.c)
        self.t.onTick(self.c)
        self.t.onTick(self.c)
        self.assertEqual(bt.Node.S_RUNNING, s.status)   # still running
        self.assertEqual(['a', 'a', 'b', 'b', 'a', 'a', 'b', 'b'], [n.name for n in self.onTickCalls])
        self.assertEqual(['a', 'b', 'a', 'b'], [n.name for n in self.onStartCalls])

    def testLoopWithFailure(self):
        """testLoopWithFailure: loop node with failure"""
        s = self.t.addNode(
            bt.Loop('l',
                bt.Sequence('s', [
                    TestAction('a', 2, bt.Node.S_FAILURE),
                    TestAction('b', 2),
                ])
            )
        )
        self.t.onTick(self.c)
        self.t.onTick(self.c)
        self.assertEqual(bt.Node.S_FAILURE, s.status)
        self.assertEqual(['a', 'a'], [n.name for n in self.onTickCalls])
        self.assertEqual(['a'], [n.name for n in self.onStartCalls])

    def testSuccessDecoratorSuccess(self):
        """testSuccessDecoratorSuccess: test of success decorator with success finish"""
        n = self.t.addNode(
            bt.AlwaysSucceed(
                'a', TestAction('a', 2)
            )
        )
        self.t.onTick(self.c)
        self.assertEqual(bt.Node.S_RUNNING, n.status)
        self.assertEqual(bt.Node.S_RUNNING, n.node.status)
        self.t.onTick(self.c)
        self.assertEqual(bt.Node.S_SUCCESS, n.status)
        self.assertEqual(bt.Node.S_SUCCESS, n.node.status)

    def testSuccessDecoratorFailure(self):
        """testSuccessDecoratorFailure: test of success decorator with failure finish"""
        n = self.t.addNode(
            bt.AlwaysSucceed(
                'a', TestAction('a', 2, bt.Node.S_FAILURE)
            )
        )
        self.t.onTick(self.c)
        self.assertEqual(bt.Node.S_RUNNING, n.status)
        self.assertEqual(bt.Node.S_RUNNING, n.node.status)
        self.t.onTick(self.c)
        self.assertEqual(bt.Node.S_SUCCESS, n.status)
        self.assertEqual(bt.Node.S_FAILURE, n.node.status)

    def testFailureDecoratorSuccess(self):
        """testFailureDecoratorSuccess: test of failure decorator with success finish"""
        n = self.t.addNode(
            bt.AlwaysFail(
                'a', TestAction('a', 2)
            )
        )
        self.t.onTick(self.c)
        self.assertEqual(bt.Node.S_RUNNING, n.status)
        self.assertEqual(bt.Node.S_RUNNING, n.node.status)
        self.t.onTick(self.c)
        self.assertEqual(bt.Node.S_FAILURE, n.status)
        self.assertEqual(bt.Node.S_SUCCESS, n.node.status)

    def testFailureDecoratorFailure(self):
        """testFailureDecoratorFailure: test of failure decorator with failure finish"""
        n = self.t.addNode(
            bt.AlwaysFail(
                'a', TestAction('a', 2, bt.Node.S_FAILURE)
            )
        )
        self.t.onTick(self.c)
        self.assertEqual(bt.Node.S_RUNNING, n.status)
        self.assertEqual(bt.Node.S_RUNNING, n.node.status)
        self.t.onTick(self.c)
        self.assertEqual(bt.Node.S_FAILURE, n.status)
        self.assertEqual(bt.Node.S_FAILURE, n.node.status)


class TestNodeClasses(unittest.TestCase):
    """Tests for the Node classes"""

    def setUp(self):
        """Set up the tests"""

    def tearDown(self):
        """Tear down the tests"""

    def testCreateNodeStatus(self):
        """testCreateNodeStatus: when create should be status None"""
        n = bt.Node('n')
        self.assertEqual(bt.Node.S_NONE, n.status)

    def testStatusChangeOnStartup(self):
        """testStatusChangeOnStartup: status should change when calling onStart"""
        n = bt.Node('n')
        n.onStart()
        self.assertEqual(bt.Node.S_SUCCESS, n.status)

    def testCanCallReset(self):
        """testCanCallReset: should be able to reset a node"""
        n = bt.Node('n')
        n.onStart()
        n.resetNode()
        self.assertEqual(bt.Node.S_NONE, n.status)


    def testResetCallsChildrenOfIterator(self):
        """testResetCallsChildrenOfIterator: reset should call the children of the iterator"""
        n = bt.Sequence('s', [
            TestAction('a', 1),
            TestAction('a', 1),
            TestAction('a', 1),
        ])
        self.assertListEqual([bt.Node.S_NONE, bt.Node.S_NONE, bt.Node.S_NONE],
                             [i.status for i in n.nodes])
        n.onTick({})
        self.assertListEqual([bt.Node.S_SUCCESS, bt.Node.S_SUCCESS, bt.Node.S_SUCCESS],
                             [i.status for i in n.nodes])
        n.resetNode()
        self.assertListEqual([bt.Node.S_NONE, bt.Node.S_NONE, bt.Node.S_NONE],
                             [i.status for i in n.nodes])
        n.onTick({})
        self.assertListEqual([bt.Node.S_SUCCESS, bt.Node.S_SUCCESS, bt.Node.S_SUCCESS],
                             [i.status for i in n.nodes])

    def testResetCallsChildOfDecorator(self):
        """testResetCallsChildOfDecorator: rest should call children of decorators"""
        n = bt.Decorator('d', TestAction('a', 1))
        self.assertEqual(bt.Node.S_NONE, n.status)
        self.assertEqual(bt.Node.S_NONE, n.node.status)
        n.onTick({})
        self.assertEqual(bt.Node.S_SUCCESS, n.status)
        self.assertEqual(bt.Node.S_SUCCESS, n.node.status)
        n.resetNode()
        self.assertEqual(bt.Node.S_NONE, n.status)
        self.assertEqual(bt.Node.S_NONE, n.node.status)


class TestNode(bt.Node):
    """A simple testing node"""

    def __init__(self, name):
        """Initialise the node"""
        super(TestNode, self).__init__(name)

    def onStart(self):
        """Called when the node is initialising"""
        super(TestNode, self).onStart()
        TestBehaviourTrees.onStartCalls.append(self)

    def onTick(self, context):
        """Called when the node is ticked"""
        super(TestNode, self).onTick(context)
        TestBehaviourTrees.onTickCalls.append(self)


class TestAction(bt.Action):
    """A simple action node"""

    def __init__(self, name, times, end=bt.Node.S_SUCCESS):
        """Initialise the action"""
        super(TestAction, self).__init__(name)
        self.times = self.original_times = times
        self.S_SUCCESS = end

    def onTick(self, context):
        """Tick the action"""
        super(TestAction, self).onTick(context)
        TestBehaviourTrees.onTickCalls.append(self)

    def doAction(self, context):
        """Perform the action"""
        for i in range(self.times - 1):
            yield self.S_RUNNING

    def onStart(self):
        """Called when the node is initialising"""
        super(TestAction, self).onStart()
        TestBehaviourTrees.onStartCalls.append(self)
        self.times = self.original_times


class OneStepAction(bt.Action):
    """A one step action"""

    def doAction(self, context):
        """Do the action"""
        TestBehaviourTrees.onTickCalls.append(self)


if __name__ == '__main__':
    unittest.main()
