"""Tests for the cards block"""

import unittest
import time

from serge.blocks import cards

from helper import *


class TestCards(unittest.TestCase, VisualTester):
    """Tests for the Cards"""

    def setUp(self):
        """Set up the tests"""
        self.d = cards.Deck()
        self.c1 = cards.Card('AD')
        self.c2 = cards.Card('AH')
        self.c3 = cards.Card('AS')
        self.c4 = cards.Card('AC')
        self.card_list = [self.c1, self.c2, self.c3, self.c4]

    def tearDown(self):
        """Tear down the tests"""

    def testCanAddCardsToDeck(self):
        """testCanAddCardsToDeck: should be able to add cards to deck"""
        self.d.addCardToTop(self.c1)
        self.d.addCardToTop(self.c2)
        self.d.addCardToTop(self.c3)
        self.d.addCardToTop(self.c4)
        self.assertEqual(4, self.d.numberOfCards())
        for c in self.card_list:
            self.assertTrue(self.d.containsCard(c))

    def testCanAddANumberOfCards(self):
        """testCanAddANumberOfCards: should be able to add a number of cards"""
        self.d.addCardsToTop(self.card_list)
        self.assertEqual(4, self.d.numberOfCards())
        for c in self.card_list:
            self.assertTrue(self.d.containsCard(c))

    def testCanWithdrawCard(self):
        """testCanWithdrawCard: should be able to withdraw a card"""
        self.d.addCardToTop(self.c1)
        self.d.addCardToTop(self.c2)
        self.d.addCardToTop(self.c3)
        self.d.addCardToTop(self.c4)
        for c in reversed(self.card_list):
            self.assertEqual(c, self.d.drawTopCard())
        self.assertEqual(0, self.d.numberOfCards())

    def testCanWithdrawMultipleCards(self):
        """testCanWithdrawMultipleCards: should be able to withdraw multiple cards"""
        self.d.addCardToTop(self.c1)
        self.d.addCardToTop(self.c2)
        self.d.addCardToTop(self.c3)
        self.d.addCardToTop(self.c4)
        got = self.d.drawTopCards(4)
        for c, g in zip(reversed(self.card_list), got):
            self.assertEqual(c, g)
        self.assertEqual(0, self.d.numberOfCards())

    def testCanWithdrawSpecificCard(self):
        """testCanWithdrawSpecificCard: should be able to withdraw a specific card"""
        self.d.addCardToTop(self.c1)
        self.d.addCardToTop(self.c2)
        self.d.addCardToTop(self.c3)
        self.d.addCardToTop(self.c4)
        g = self.d.drawSpecificCard(self.c2)
        self.assertEqual(3, self.d.numberOfCards())
        self.assertEqual(g, self.c2)
        self.assertFalse(self.d.containsCard(self.c2))

    def testFailWithdrawEmpty(self):
        """testFailWithdrawEmpty: should fail when withdrawing from an empty deck"""
        self.assertRaises(cards.NoCard, self.d.drawTopCard)

    def testFailWithdrawMissingSpecific(self):
        """testFailWithdrawMissingSpecific: should fail with withdrawing a specific card that is not there"""
        self.d.addCardToTop(self.c1)
        self.d.addCardToTop(self.c2)
        self.d.addCardToTop(self.c3)
        self.assertRaises(cards.CardNotFound, self.d.drawSpecificCard, self.c4)

    def testCanShuffle(self):
        """testCanShuffle: should be able to shuffle a deck"""
        self.d.addCardsToTop(self.card_list)
        self.assertListEqual(self.card_list, self.d.getCards())
        self.d.shuffleCards()
        self.assertNotEqual(self.card_list, self.d.getCards())

    def testCanClear(self):
        """testCanClear: should be able to clear a deck"""
        self.d.addCardsToTop(self.card_list)
        self.d.clearCards()
        self.assertEqual(0, self.d.numberOfCards())

    def testCanGetStandardDeck(self):
        """testCanGetStandardDeck: should be able to get a standard deck"""
        d = cards.StandardDeck()
        self.assertEqual(52, d.numberOfCards())


if __name__ == '__main__':
    unittest.main()
