"""Tests for the cellular automation block"""

import unittest

import serge.render
import serge.engine
import serge.actor
import serge.visual
import serge.world
import serge.events
import serge.blocks.utils
import serge.blocks.cellular

from helper import *


class TestCellular(unittest.TestCase, VisualTester):
    """Tests for the Cellular"""

    def setUp(self):
        """Set up the tests"""
        self.c = serge.blocks.cellular.Otomata('a', 'a', (3, 3))
        self.bounces = []
        
    def tearDown(self):
        """Tear down the tests"""

    def testMovement(self):
        """testMovement: should be able to see movement in all four directions"""
        for (state, new_cell) in ((self.c.S_UP, (1, 0)), (self.c.S_DOWN, (1, 2)),
                                  (self.c.S_LEFT, (0, 1)), (self.c.S_RIGHT, (2, 1))):
            self.c.resetAutomaton()
            self.c.setState((1, 1), state)
            self.c.updateState()
            #
            # State of origin cell should be null
            self.assertEqual([self.c.S_NULL], self.c.getState((1, 1)))
            #
            # New cell location should have changed, but only this one
            for dx in (-1, 0, +1):
                for dy in (-1, 0, +1):
                    cx, cy = 1 + dx, 1 + dy
                    if (cx, cy) == new_cell:
                        self.assertEqual([state], self.c.getState((cx, cy)))
                    else:
                        self.assertEqual([self.c.S_NULL], self.c.getState((cx, cy)))

    def testBouncing(self):
        """testBouncing: should be able to see bouncing in all four directions"""
        for (state, new_state, new_cell) in ((self.c.S_UP, self.c.S_DOWN, (1, 0)),
                                             (self.c.S_DOWN, self.c.S_UP, (1, 2)),
                                             (self.c.S_LEFT, self.c.S_RIGHT, (0, 1)),
                                             (self.c.S_RIGHT, self.c.S_LEFT, (2, 1))):
            self.c.resetAutomaton()
            self.c.setState((1, 1), state)
            self.c.updateState()
            self.c.updateState()
            #
            # State of origin cell should be null
            self.assertEqual([self.c.S_NULL], self.c.getState((1, 1)))
            #
            # New cell location should have changed, but only this one
            for dx in (-1, 0, +1):
                for dy in (-1, 0, +1):
                    cx, cy = 1 + dx, 1 + dy
                    if (cx, cy) == new_cell:
                        self.assertEqual([new_state], self.c.getState((cx, cy)))
                    else:
                        self.assertEqual([self.c.S_NULL], self.c.getState((cx, cy)))
    
    def testInteractionTurnsClockwise(self):
        """testInteractionTurnsClockwise: should be able to see cells interacting to change state"""
        self.c.setState((0, 1), self.c.S_RIGHT)
        self.c.setState((1, 2), self.c.S_UP)
        #
        self.c.updateState()
        #
        # Middle square should be combination of items
        self.assertSetEqual(set([self.c.S_DOWN, self.c.S_RIGHT]), set(self.c.getState((1, 1))))
        #
        self.c.updateState()
        #
        # Should have now resolved into two separate ones again
        self.assertEqual([self.c.S_NULL], self.c.getState((1, 1)))
        self.assertEqual([self.c.S_NULL], self.c.getState((0, 1)))
        self.assertEqual([self.c.S_NULL], self.c.getState((1, 0)))
        self.assertEqual([self.c.S_RIGHT], self.c.getState((2, 1)))
        self.assertEqual([self.c.S_DOWN], self.c.getState((1, 2)))

    def testCallbackOnEdges(self):
        """testCallbackOnEdges: should get callback on edge hits"""
        self.c.linkEvent(self.c.E_BOUNCE, self._bounceCallback)
        for (state, new_cell) in ((self.c.S_UP, (1, -1)),
                                  (self.c.S_DOWN, (1, 3)),
                                  (self.c.S_LEFT, (-1, 1)),
                                  (self.c.S_RIGHT, (3, 1))):
            self.c.resetAutomaton()
            self.bounces = []
            self.c.setState((1, 1), state)
            #
            # No bounces from a single move
            self.c.updateState()
            self.assertListEqual([], self.bounces)
            #
            # Bounce should occur now
            self.c.updateState()
            self.assertListEqual([new_cell], self.bounces)

    def _bounceCallback(self, obj, arg):
        """A bounce occurred"""
        self.bounces.append(obj)

    def testFailSetStateOutOfBounds(self):
        """testFailSetStateOutOfBounds: should fail when setting state out of bounds"""
        for cell in ((-1, 0), (0, -1), (-1, -1), (3, 1), (1, 3), (3, 3)):
            self.assertRaises(serge.blocks.cellular.OutOfRange, self.c.setState, cell, self.c.S_DOWN)
    
    def testFailGetStateOutOfBounds(self):
        """testFailGetStateOutOfBounds: should fail when getting state out of bounds"""
        for cell in ((-1, 0), (0, -1), (-1, -1), (3, 1), (1, 3), (3, 3)):
            self.assertRaises(serge.blocks.cellular.OutOfRange, self.c.getState, cell)
    
    def testCanSetUpdateFrequency(self):
        """testCanSetUpdateFrequency: should be able to set the update frequency"""
        self.c.setFrequency(500.0)
        self.c.linkEvent(self.c.E_BOUNCE, self._bounceCallback)
        self.c.setState((1, 1), self.c.S_UP)
        #
        self.c.updateActor(500, None)
        self.assertListEqual([], self.bounces)
        self.c.updateActor(499, None)
        self.assertListEqual([], self.bounces)
        self.c.updateActor(2, None)
        self.assertListEqual([(1, -1)], self.bounces)


if __name__ == '__main__':
    unittest.main()