"""Tests for Geometry stuff"""

import math
import random
import unittest
import serge.blocks.hexgrid


class TestHexGrid(unittest.TestCase):
    """Test for the hexagonal grid"""

    def setUp(self):
        """Set up the test cases"""
        self.longMessage = True
        self.g = serge.blocks.hexgrid.HexagonalGrid(6, 4)

    def tearDown(self):
        """Tear down the test cases"""

    def testCanCreate(self):
        """testCanCreate: should be bale to create a grid"""

    def _cellId(self, x, y):
        """Return a cell ID"""
        return '%s-%s' % (x, y)

    def _setIds(self):
        """Set the ids of the grid cells"""
        for x in range(self.g.width):
            for y in range(self.g.height):
                self.g.setCell(x, y, self._cellId(x, y))

    def testShouldBeEmpty(self):
        """testShouldBeEmpty: grid should be empty initially"""
        for x in range(6):
            for y in range(4):
                self.assertEqual(None, self.g.getCell(x, y))

    def testCanSetAndGetCells(self):
        """testCanSetAndGetCells: should be able to get and set cells"""
        self._setIds()
        #
        for x in range(6):
            for y in range(4):
                self.assertEqual(self._cellId(x, y), self.g.getCell(x, y))


    def testGridExtent(self):
        """testGridExtent: should have consistent grid extent"""
        for x, y in (-1, 0), (0, -1), (6, 0), (0, 4):
            self.assertRaises(serge.blocks.hexgrid.OutOfRange, self.g.getCell, x, y)
            self.assertRaises(serge.blocks.hexgrid.OutOfRange, self.g.setCell, x, y, None)
            self.assertRaises(serge.blocks.hexgrid.OutOfRange, self.g.getNeighbourLocations, x, y)
            self.assertFalse(self.g.cellIsInGrid(x, y))

    def testCanGetNeighbours(self):
        """testCanGetNeighbours: should be able to get neighbour cells"""
        tests = [
            [
                (1, 1),
                [
                    (0, 1), (1, 0), (2, 1), (0, 2), (1, 2), (2, 2),
                ]
            ],
            [
                (3, 2),
                [
                    (2, 2), (3, 1), (4, 2), (2, 3), (3, 3), (4, 3),
                ]
            ],
            [
                (4, 2),
                [
                    (3, 1), (4, 1), (5, 1), (3, 2), (4, 3), (5, 2),
                ]
            ],
            [
                (0, 0),
                [
                    (0, 1), (1, 0),
                ]
            ],
            [
                (6, 4),
                [
                    (5, 3), (6, 3), (5, 4),
                ]
            ],
            [
                (0, 2),
                [
                    (0, 1), (1, 1), (0, 3), (1, 2),
                ]
            ],
            [
                (0, 4),
                [
                    (0, 3), (1, 3), (1, 4),
                ]
            ],
            [
                (6, 0),
                [
                    (5, 0), (6, 1),
                ]
            ],
            [
                (6, 2),
                [
                    (5, 1), (6, 1), (5, 2), (6, 3),
                ]
            ],
        ]
        self.g = serge.blocks.hexgrid.HexagonalGrid(7, 5)
        self._setIds()
        #
        for (x, y), neighbours in tests:
            neighbour_set = set(neighbours)
            grid_neighbour_locations = set(self.g.getNeighbourLocations(x, y))
            self.assertSetEqual(neighbour_set, grid_neighbour_locations, 'Test for cell %s, %s' % (x, y))
            #
            # Try to get the neighbours directly
            neighbour_ids = set(self._cellId(cx, cy) for cx, cy in neighbours)
            got_ids = set(cell for cell in self.g.getNeighbours(x, y))
            self.assertSetEqual(neighbour_ids, got_ids, 'Failed on cell %s, %s' % (x, y))

    def testCanGetCellsWithinDistance(self):
        """testCanGetCellsWithinDistance: should be able to get all cells within a certain distance"""
        tests = [
            # Tests around 0, 0
            ((0, 0), 1, [
                (1, 0), (0, 1)
            ]),
            ((0, 0), 2, [
                (1, 0), (0, 1),
                (2, 0), (2, 1), (1, 1), (0, 2),
            ]),
            ((0, 0), 3, [
                (1, 0), (0, 1),
                (2, 0), (2, 1), (1, 1), (0, 1),
                (3, 0), (3, 1), (2, 2), (1, 2), (0, 2), (0, 3)
            ]),

            # Tests around 2, 3
            ((2, 3), 1, [
                (1, 2), (2, 2), (3, 2), (3, 3), (2, 4), (1, 3),
            ]),
            ((2, 3), 2, [
                (1, 2), (2, 2), (3, 2), (3, 3), (2, 4), (1, 3),
                (0, 3), (0, 2), (1, 1), (2, 1), (3, 1), (4, 2), (4, 3), (4, 4), (3, 4), (1, 4), (0, 4),
            ]),
            ((2, 3), 3, [
                (1, 2), (2, 2), (3, 2), (3, 3), (2, 4), (1, 3),
                (0, 3), (0, 2), (1, 1), (2, 1), (3, 1), (4, 2), (4, 3), (4, 4), (3, 4), (1, 4), (0, 4),
                (0, 1), (1, 0), (2, 0), (3, 0), (4, 1), (5, 1), (5, 2), (5, 2), (5, 3), (5, 4),
            ]),

            # Tests around 1, 4
            ((1, 4), 1, [
                (0, 4), (1, 3), (2, 4),
            ]),
            ((1, 4), 2, [
                (0, 4), (1, 3), (2, 4),
                (0, 3), (1, 2), (2, 3), (3, 3), (3, 4),
            ]),
            ((1, 4), 3, [
                (0, 4), (1, 3), (2, 4),
                (0, 3), (1, 2), (2, 3), (3, 3), (3, 4),
                (0, 2), (1, 1), (2, 2), (3, 2), (4, 3), (4, 4),
            ]),
        ]
        self.g = serge.blocks.hexgrid.HexagonalGrid(7, 5)
        self._setIds()
        for (cx, cy), distance, expected_neighbours in tests:
            #
            # Do for cell locations
            neighbours = self.g.getNeighbourLocationsWithin(cx, cy, distance)
            self.assertSetEqual(set(expected_neighbours), set(neighbours),
                                'Locations miss for %s, distance %s' % ((cx, cy), distance))
            #
            # Now repeat for cells
            neighbour_cells = self.g.getNeighbourCellsWithin(cx, cy, distance)
            expected = [self._cellId(tx, ty) for tx, ty in expected_neighbours]
            self.assertSetEqual(set(expected), set(neighbour_cells),
                                'Cell miss for %s, distance %s' % ((cx, cy), distance))

    def testCanGetRelativeCells(self):
        """testCanGetRelativeCells: should be able to get cells in different directions"""
        tests = [
            [
                (1, 1),
                [
                    ('n', (1, 0)),
                    ('s', (1, 2)),
                    ('nw', (0, 1)),
                    ('ne', (2, 1)),
                    ('sw', (0, 2)),
                    ('se', (2, 2)),
                ],
            ],
            [
                (2, 1),
                [
                    ('n', (2, 0)),
                    ('s', (2, 2)),
                    ('nw', (1, 0)),
                    ('ne', (3, 0)),
                    ('sw', (1, 1)),
                    ('se', (3, 1)),
                ],
            ],
        ]
        self._setIds()
        #
        for (x, y), neighbours in tests:
            for direction, (cx, cy) in neighbours:
                self.assertEqual((cx, cy), self.g.getRelativeCellCoords(x, y, direction))
                self.assertEqual(self._cellId(cx, cy), self.g.getRelativeCell(x, y, direction))

    def testRelativeCellsOutOfRange(self):
        """testRelativeCellsOutOfRange: relative cells out of range should raise exception"""
        tests = [
            [
                (0, 0),
                [
                    ['n', 'nw', 'ne', 'sw'],
                    ['se', 's'],
                ]
            ],
            [
                (5, 3),
                [
                    ['sw', 's', 'se', 'ne'],
                    ['n', 'nw'],
                ]
            ],
            [
                (2, 0),
                [
                    ['nw', 'n', 'ne'],
                    ['sw', 's', 'se'],
                ]
            ],
            [
                (2, 3),
                [
                    ['s'],
                    ['sw', 'nw', 'n', 'ne', 'se']
                ]
            ],
        ]
        for (x, y), (out_of_ranges, in_ranges) in tests:
            for direction in out_of_ranges:
                self.assertRaises(serge.blocks.hexgrid.OutOfRange, self.g.getRelativeCellCoords, x, y, direction)
            for direction in in_ranges:
                self.g.getRelativeCellCoords(x, y, direction)

    def testCanGetDirections(self):
        """testCanGetDirections: should be able to get direction of cell from another"""
        for x in range(6):
            for y in range(4):
                neighbours = self.g.getNeighbourLocations(x, y)
                for cx, cy in neighbours:
                    direction = self.g.getCellDirection((cx, cy), (x, y))
                    nx, ny = self.g.getRelativeCellCoords(x, y, direction)
                    self.assertEqual((cx, cy), (nx, ny),
                                     'Cell %s, %s other %s, %s. Direction %s' % (x, y, cx, cy, direction))

    def testCellDirectionFailsIfTooFar(self):
        """testCellDirectionFailsIfTooFar: testing for cell direction must be a neighbour"""
        tests = [
            [(1, 1), (3, 2)],
            [(1, 1), (1, 3)],
            [(1, 1), (4, 0)],
        ]
        for (x, y), (cx, cy) in tests:
            self.assertRaises(serge.blocks.hexgrid.OutOfRange, self.g.getCellDirection, (x, y), (cx, cy))

    def testGetCellLocation(self):
        """testGetCellLocation: should be able to get cell location"""
        h = math.sqrt(3.0 / 4.0)
        tests = [
            [(0, 0), (1.0, h)],
            [(1, 0), (2.5, h * 2)],
            [(2, 0), (4.0, h)],
            [(3, 0), (5.5, h * 2)],

            [(0, 1), (1.0, h * 3)],
            [(1, 1), (2.5, h * 4)],
            [(2, 1), (4.0, h * 3)],
            [(3, 1), (5.5, h * 4)],
        ]
        for (x, y), (px, py) in tests:
            self.assertEqual((px, py), self.g.getCellLocation(x, y), 'Cell %s, %s' % (x, y))

    def testGetCellPoints(self):
        """testGetCellPoints: should be able to get the cell points"""
        tests = [
            [
                (0, 0),
                [
                    (0.5, 0.0),
                    (1.5, 0.0),
                    (2.0, self.g.half_size),
                    (1.5, 2 * self.g.half_size),
                    (0.5, 2 * self.g.half_size),
                    (0.0, self.g.half_size),
                ]
            ]
        ]
        for (x, y), points in tests:
            cell_points = self.g.getCellPoints(x, y)
            for (px, py), (tx, ty) in zip(points, cell_points):
                self.assertEqual((px, py), (tx, ty), 'Failed on cell %s, %s' % (x, y))

    def testGetCellFromLocation(self):
        """testGetCellFromLocation: should be able to get a cell from a location"""
        #
        # Store centers
        centers = []
        for x in range(6):
            for y in range(4):
                centers.append((self.g.getCellLocation(x, y), (x, y)))
        #
        self._setIds()
        #
        ITERATIONS = 100
        for _ in range(ITERATIONS):
            x = random.uniform(0.3, 9.0)
            y = random.uniform(math.sqrt(3.0 / 4.0), 6.5)
            px, py = self._getClosest(x, y, centers)
            cx, cy = self.g.getCellCoordsContaining(x, y)
            self.assertEqual((px, py), (cx, cy), 'Failed on points %s, %s' % (x, y))
            #
            cell = self.g.getCellContaining(x, y)
            self.assertEqual(self._cellId(px, py), cell, 'Failed on points %s, %s' % (x, y))

    def testCellLocationOutOfRange(self):
        """testCellLocationOutOfRange: should raise an error if getting a cell that is out of range"""
        centers = [
            (0, 0),
            (-10, -10),
            (1000, 1000),
            (2, 0),
            (0, 2),
        ]
        for px, py in centers:
            self.assertRaises(serge.blocks.hexgrid.OutOfRange, self.g.getCellContaining, px, py)

    def _getClosest(self, x, y, centers):
        """Return the closest cell to the center"""
        distances = []
        for (px, py), (cx, cy) in centers:
            distance = math.sqrt((x - px) ** 2 + (y - py) ** 2)
            distances.append((distance, (cx, cy)))
        distances.sort()
        return distances[0][1]

    def testGetCellDistances(self):
        """testGetCellDistances: should be able to get distances from one cell to another"""
        tests = [
            [
                (2, 2),
                [
                    ((2, 2), 0),
                    ((1, 1), 1),
                    ((3, 2), 1),
                    ((2, 3), 1),
                    ((1, 0), 2),
                    ((2, 4), 2),
                    ((3, 3), 2),
                    ((5, 3), 3),
                ],
            ],
            [
                (3, 3),
                [
                    ((3, 3), 0),
                    ((3, 2), 1),
                    ((4, 3), 1),
                    ((2, 3), 1),
                    ((3, 1), 2),
                    ((2, 2), 2),
                    ((1, 3), 2),
                    ((2, 1), 3),
                ],
            ],
            [
                (0, 0),
                [
                    ((0, 2), 2),
                    ((2, 3), 4),
                    ((4, 0), 4),
                    ((4, 1), 4),
                ]
            ],
            [
                (2, 3),
                [
                    ((3, 2), 1),
                    ((2, 4), 1),
                ],
            ],
        ]
        self.g.getCellDistance((2, 3), (3, 2))
        for (x, y), items in tests:
            for (cx, cy), distance in items:
                self.assertEqual(distance, self.g.getCellDistance((x, y), (cx, cy)),
                                 'Failed %s : %s' % ((x, y), (cx, cy)))
                self.assertEqual(distance, self.g.getCellDistance((cx, cy), (x, y)),
                                 'Failed %s : %s' % ((cx, cy), (x, y)))

    def testCanIterateOnLocations(self):
        """testCanIterateOnLocations: should be able to iterate on cell locations"""
        locations = set()
        for loc in self.g.iterLocations():
            locations.add(loc)
        #
        expected = set()
        for x in range(6):
            for y in range(4):
                expected.add((x, y))
        #
        self.assertSetEqual(expected, locations)

    def testCanIterateOnCells(self):
        """testCanIterateOnCells: should be able to iterate on cells"""
        self._setIds()
        cells = set()
        for loc in self.g.iterCells():
            cells.add(loc)
        #
        expected = set()
        for x in range(6):
            for y in range(4):
                expected.add(self._cellId(x, y))
        #
        self.assertSetEqual(expected, cells)


if __name__ == '__main__':
    unittest.main()
