"""Tests for the Text Generator part"""

import unittest
import time

import serge.render
import serge.engine
import serge.actor
import serge.visual
import serge.world
import serge.events
import serge.blocks.utils
import serge.blocks.textgenerator

from helper import *


class TestTextGenerator(unittest.TestCase, VisualTester):
    """Tests for the TextGenerator"""
    
    def setUp(self):
        """Set up the tests"""
        self.t = serge.blocks.textgenerator.TextGenerator()
        self.e1 = '''
                        colour: red
                        colour: blue
                        colour: green
                        name: bob
                        name: fred
                        thing: this
                        size: big
                        size: small
                        colourname: @{colour}@ @{name}@
                        cn: @{colourname}@
                    '''
        self.e2 = '''
                        colour {
                        red
                        blue
                        green
                        }
                        name {
                        bob
                        fred
                        }
                        thing: this
                        size: big
                        size: small
                        colourname: @{colour}@ @{name}@
                        cn: @{colourname}@
                  '''
        self.e3 = '''
                        sex: male
                        sex: female
                        male-name: bob
                        female-name: alice
                        name: @{@{sex}@-name}@
                  '''
                    
    def tearDown(self):
        """Tear down the tests"""
    
    def testCanCreate(self):
        """testCanCreate: should be able to create the generator"""
        self.assert_(self.t is not None)
    
    def testCanAddExamplesByMethod(self):
        """testCanAddExamplesByMethod: should be able to add examples by using the methods"""
        self.t.addExample('colour', 'red')
        self.assertEqual('red', self.t.getRandomFormCompletion('colour'))
        
    def testCanAddExampleFromText(self):
        """testCanAddExampleFromText: should be able to add examples from text"""
        self.t.addExamplesFromText('colour: red\nname: bob\n\n\tthing : this')
        self.assertEqual('red', self.t.getRandomFormCompletion('colour'))
        self.assertEqual('bob', self.t.getRandomFormCompletion('name'))
        self.assertEqual('this', self.t.getRandomFormCompletion('thing'))

    def testMultiExampleText(self):
        """testMultiExampleText: should be able to use multiexamples in text"""
        self.t.addExamplesFromText(self.e2)
        s = self.t.getRandomSentence('@{colour}@ @{name}@')
        a, b = s.split(' ')
        self.assertEqual(True, a in ['red', 'blue', 'green'])
        self.assertEqual(True, b in ['bob', 'fred'])
    
    def testSingleReplacement(self):
        """testSingleReplacement: should be able to do a single replacement"""
        self.t.addExamplesFromText(self.e1)
        s = self.t.getRandomSentence('@{thing}@')
        self.assertEqual('this', s)
        
    def testSingleOneFromMany(self):
        """testSingleOneFromMany: should be able to do a one from many replacement"""
        self.t.addExamplesFromText(self.e1)
        s = self.t.getRandomSentence('@{colour}@')
        self.assertEqual(True, s in ['red', 'blue', 'green'])
    
    def testTwoSinglesInOne(self):
        """testTwoSinglesInOne: should be able to do two examples of a single replacement"""
        self.t.addExamplesFromText(self.e1)
        s = self.t.getRandomSentence('@{colour}@ @{name}@')
        a, b = s.split(' ')
        self.assertEqual(True, a in ['red', 'blue', 'green'])
        self.assertEqual(True, b in ['bob', 'fred'])
    
    def testDoubleLevelReplacement(self):
        """testDoubleLevelReplacement: should be able to do a two level replacement"""
        self.t.addExamplesFromText(self.e1)
        s = self.t.getRandomSentence('@{colourname}@')
        a, b = s.split(' ')
        self.assertEqual(True, a in ['red', 'blue', 'green'])
        self.assertEqual(True, b in ['bob', 'fred'])
    
    def testRecursizeRepalcement(self):
        """testRecursizeRepalcement: should be able to do recursive replacement"""
        self.t.addExamplesFromText(self.e1)
        s = self.t.getRandomSentence('@{cn}@')
        a, b = s.split(' ')
        self.assertEqual(True, a in ['red', 'blue', 'green'])
        self.assertEqual(True, b in ['bob', 'fred'])
    
    def testNestedReplacement(self):
        """testNestedReplacement: should be able to do a nested replacement"""
        self.t.addExamplesFromText(self.e3)
        self.assertEqual('bob', self.t.getRandomSentence('@{name}@', {'sex':'male'}))
        self.assertEqual('alice', self.t.getRandomSentence('@{name}@', {'sex':'female'}))
    
    def testCanCheckPropertiesAfterwards(self):
        """testCanCheckPropertiesAfterwards: shold be able to check properties afterwars"""
        self.t.addExamplesFromText(self.e3)
        p = {'sex':'male'}
        self.assertEqual('bob', self.t.getRandomSentence('@{name}@', p))
        self.assertEqual('bob', self.t.getRandomSentence('@{name}@', p))


# Some facts
F_TEST = 'test-fact'
F_SCORE = 'score-small-fact'
F_BIG_SCORE = 'score-big-fact'
F_ALL_SCORE = 'score-*-fact'

R_SCORED = 'someone-scored'
R_SCORED_1 = 'someone-scored-1'
R_SCORED_2 = 'someone-scored-2'
R_SCORED_VERY_RECENTLY = 'someone-scored-very-recently'
R_SCORED_RECENTLY = 'someone-scored-recently'


class TestFactDatabase(unittest.TestCase, VisualTester):
    """Tests for the RuleSystem"""

    def setUp(self):
        """Set up the tests"""
        self.db = serge.blocks.textgenerator.FactDatabase()

    def tearDown(self):
        """Tear down the tests"""

    def testCanCreateFactDatabase(self):
        """testCanCreateFactDatabase: should be able to create a fact database"""

    def testCanAddFact(self):
        """testCanAddFact: should be able to add a fact to db"""
        self.db.addFact(F_TEST)

    def testCanControlTimeForDatabase(self):
        """testCanControlTimeForDatabase: should be able to control the time for the database"""
        self.assertEqual(0, self.db.getTime())
        self.db.setTime(100)
        self.assertEqual(100, self.db.getTime())

    def testFailSetTimeEarlier(self):
        """testFailSetTimeEarlier: should fail if try to go back in time"""
        self.db.setTime(100)
        # Can duplicate
        self.db.setTime(100)
        self.assertRaises(serge.blocks.textgenerator.InvalidTime, self.db.setTime, 99)

    def testFactShouldHaveOccurredTime(self):
        """testFactShouldHaveOccurredTime: should be able to have occurrence time for fact"""
        self.db.setTime(100)
        self.db.addFact(F_TEST)
        self.db.setTime(150)
        self.assertEqual(100, self.db.getLastOccurrenceTime(F_TEST))

    def testCanCheckIfFactOccurred(self):
        """testCanCheckIfFactOccurred: should be able to check if fact occurred"""
        self.assertFalse(self.db.hasOccurred(F_TEST))
        self.assertFalse(self.db.hasOccurred(F_ALL_SCORE))
        self.db.addFact(F_TEST)
        self.assertTrue(self.db.hasOccurred(F_TEST))

    def testFailGetOccurrenceTimeWhenNoFact(self):
        """testFailGetOccurrenceTimeWhenNoFact: should raise an error when checking occurrence time if not occurred"""
        self.assertRaises(serge.blocks.textgenerator.NoOccurrence, self.db.getLastOccurrenceTime, F_TEST)

    def testOccurrenceTimeShouldReflectLatest(self):
        """testOccurrenceTimeShouldReflectLatest: testing the occurrence time should reflect the latest"""
        self.db.setTime(100)
        self.db.addFact(F_TEST)
        self.db.setTime(200)
        self.db.addFact(F_TEST)
        self.db.setTime(250)
        self.assertEqual(200, self.db.getLastOccurrenceTime(F_TEST))

    def testCanTrackNumberOfOccurrences(self):
        """testCanTrackNumberOfOccurrences: should be able to track number of times a fact occurred"""
        self.assertEqual(0, self.db.getNumberOfOccurrences(F_TEST))
        self.db.addFact(F_TEST)
        self.db.addFact(F_TEST)
        self.db.addFact(F_TEST)
        self.assertEqual(3, self.db.getNumberOfOccurrences(F_TEST))

    def testCanTackFactFrequency(self):
        """testCanTackFactFrequency: should be able to track fact frequency"""
        self.assertEqual(0, self.db.getOccurrenceFrequency(F_TEST))
        for tme in range(10):
            self.db.setTime(tme)
            self.db.addFact(F_TEST)
        self.db.setTime(10)
        self.assertEqual(1.0, self.db.getOccurrenceFrequency(F_TEST))

    def testCanAddParameterToFact(self):
        """testCanAddParameterToFact: should be able to add a parameter to a fact"""
        self.db.addFact(F_SCORE, 123)

    def testCanRetrieveFactParameter(self):
        """testCanRetrieveFactParameter: should be able to retrieve fact parameter"""
        self.db.addFact(F_SCORE, 123)
        self.db.addFact(F_SCORE, 456)
        self.assertEqual(456, self.db.getLastFactParameters(F_SCORE)[0])

    def testCanRetrieveAllFacts(self):
        """testCanRetrieveAllFactParameters: should be able to retrieve all facts"""
        for tme, param in enumerate([123, 456, 789, 1024]):
            self.db.setTime(tme)
            self.db.addFact(F_SCORE, param)
        facts = self.db.getFactInstances(F_SCORE)
        self.assertListEqual([0, 1, 2, 3], [fact[0] for fact in facts])
        self.assertListEqual([123, 456, 789, 1024], [fact[1] for fact in facts])

    def testRetrieveFactsWithNoInstances(self):
        """testRetrieveFactsWithNoInstances: retrieving all when none gives empty list"""
        self.assertListEqual([], self.db.getFactInstances(F_SCORE))

    def testCanPatternMatchEvent(self):
        """testCanPatternMatchEvent: Methods that look for events can take patterns"""
        for tme, (fact, param) in enumerate(
                [(F_SCORE, 123), (F_BIG_SCORE, 456), (F_SCORE, 789),
                 (F_BIG_SCORE, 1024)]):
            self.db.setTime(tme)
            self.db.addFact(fact, param)
        #
        self.assertTrue(self.db.hasOccurred(F_ALL_SCORE))
        self.assertTrue(self.db.hasOccurred(F_BIG_SCORE))
        self.assertTrue(self.db.hasOccurred(F_SCORE))
        #
        self.assertEqual(4, self.db.getNumberOfOccurrences(F_ALL_SCORE))
        self.assertEqual(2, self.db.getNumberOfOccurrences(F_BIG_SCORE))
        self.assertEqual(2, self.db.getNumberOfOccurrences(F_SCORE))
        #
        facts = self.db.getFactInstances(F_SCORE)
        facts_big = self.db.getFactInstances(F_BIG_SCORE)
        facts_all = self.db.getFactInstances(F_ALL_SCORE)
        #
        self.assertListEqual([0, 1, 2, 3], [fact[0] for fact in facts_all])
        self.assertListEqual([123, 456, 789, 1024], [fact[1] for fact in facts_all])
        #
        self.assertListEqual([0, 2], [fact[0] for fact in facts])
        self.assertListEqual([123, 789], [fact[1] for fact in facts])
        #
        self.assertListEqual([1, 3], [fact[0] for fact in facts_big])
        self.assertListEqual([456, 1024], [fact[1] for fact in facts_big])
        #
        self.assertEqual(3, self.db.getLastOccurrenceTime(F_ALL_SCORE))
        self.assertEqual(3, self.db.getLastOccurrenceTime(F_BIG_SCORE))
        self.assertEqual(2, self.db.getLastOccurrenceTime(F_SCORE))
        #
        self.assertEqual(1024, self.db.getLastFactParameters(F_ALL_SCORE)[0])
        self.assertEqual(1024, self.db.getLastFactParameters(F_BIG_SCORE)[0])
        self.assertEqual(789, self.db.getLastFactParameters(F_SCORE)[0])


class TestRuleSystem(unittest.TestCase, VisualTester):
    """Tests for the RuleSystem"""

    def setUp(self):
        """Set up the tests"""
        self.db = serge.blocks.textgenerator.FactDatabase()
        self.s = serge.blocks.textgenerator.RuleSystem(self.db)

    def tearDown(self):
        """Tear down the tests"""

    def testCanCreateRuleSystem(self):
        """testCanCreateRuleSystem: should be able to create a rule system"""

    def testAddASimpleRule(self):
        """testAddASimpleRule: should be able to add a simple rule"""
        self.s.addRule(serge.blocks.textgenerator.Rule(
            R_SCORED, [
                self.s.FactOccurred(F_SCORE),
            ]
        ))

    def testRuleToCheckForFact(self):
        """testRuleToCheckForFact: run checking for a fact existing"""
        self.s.addRule(serge.blocks.textgenerator.Rule(
            R_SCORED, [
                self.s.FactOccurred(F_SCORE),
            ]
        ))
        # Initially should be no match
        self.assertEqual(None, self.s.getBestMatchingRule())
        # Still should be no match if wrong rule
        self.db.addFact(F_BIG_SCORE)
        self.assertEqual(None, self.s.getBestMatchingRule())
        #
        self.db.addFact(F_SCORE)
        self.assertEqual(R_SCORED, self.s.getBestMatchingRule().rule_id)

    def testRecentFact(self):
        """testRecentFact: rule to check for recent fact"""
        self.s.addRule(serge.blocks.textgenerator.Rule(
            R_SCORED_RECENTLY, [
                self.s.FactRecentlyOccurred(F_SCORE, 10),
            ]
        ))
        # Initially should be no match
        self.assertEqual(None, self.s.getBestMatchingRule())
        # Still should be no match if wrong rule
        self.db.addFact(F_BIG_SCORE)
        self.assertEqual(None, self.s.getBestMatchingRule())
        # Rule should now match
        self.db.addFact(F_SCORE)
        self.assertEqual(R_SCORED_RECENTLY, self.s.getBestMatchingRule().rule_id)
        # But match goes away after 10s
        self.db.setTime(11)
        self.assertEqual(None, self.s.getBestMatchingRule())

    def testNumberOfFacts(self):
        """testNumberOfFacts: rule to check for a certain number of facts"""
        self.s.addRule(serge.blocks.textgenerator.Rule(
            R_SCORED, [
                self.s.FactOccurredExactNumberOfTimes(F_SCORE, 4),
            ]
        ))
        # Initially should be no match
        self.assertEqual(None, self.s.getBestMatchingRule())
        # Still should be no match if wrong rule
        self.db.addFact(F_BIG_SCORE)
        self.assertEqual(None, self.s.getBestMatchingRule())
        #
        # No match for 3 times
        self.db.addFact(F_SCORE)
        self.db.addFact(F_SCORE)
        self.db.addFact(F_SCORE)
        self.assertEqual(None, self.s.getBestMatchingRule())
        # One more and should match
        self.db.addFact(F_SCORE)
        self.assertEqual(R_SCORED, self.s.getBestMatchingRule().rule_id)
        # Add one more and should no longer match
        self.db.addFact(F_SCORE)
        self.assertEqual(None, self.s.getBestMatchingRule())

    def testAtLeastNumberOfFacts(self):
        """testAtLeastNumberOfFacts: rule to check for at least a number"""
        self.s.addRule(serge.blocks.textgenerator.Rule(
            R_SCORED, [
                self.s.FactOccurredNumberOfTimes(F_SCORE, 4),
            ]
        ))
        # Initially should be no match
        self.assertEqual(None, self.s.getBestMatchingRule())
        # Still should be no match if wrong rule
        self.db.addFact(F_BIG_SCORE)
        self.assertEqual(None, self.s.getBestMatchingRule())
        #
        # No match for 3 times
        self.db.addFact(F_SCORE)
        self.db.addFact(F_SCORE)
        self.db.addFact(F_SCORE)
        self.assertEqual(None, self.s.getBestMatchingRule())
        # One more and should match
        self.db.addFact(F_SCORE)
        self.assertEqual(R_SCORED, self.s.getBestMatchingRule().rule_id)
        # Add one more and should still match
        self.db.addFact(F_SCORE)
        self.assertEqual(R_SCORED, self.s.getBestMatchingRule().rule_id)

    def testFactFrequency(self):
        """testFactFrequency: rule to check for frequency to exceed value"""
        self.s.addRule(serge.blocks.textgenerator.Rule(
            R_SCORED, [
                self.s.FactOccurredFrequency(F_SCORE, 0.5),
            ]
        ))
        # Initially should be no match
        self.assertEqual(None, self.s.getBestMatchingRule())
        # Still should be no match if wrong rule
        self.db.addFact(F_BIG_SCORE)
        self.assertEqual(None, self.s.getBestMatchingRule())
        #
        # No match for 4 times in 10 s
        self.db.setTime(10)
        self.db.addFact(F_SCORE)
        self.db.addFact(F_SCORE)
        self.db.addFact(F_SCORE)
        self.db.addFact(F_SCORE)
        self.assertEqual(None, self.s.getBestMatchingRule())
        # One more and should match
        self.db.addFact(F_SCORE)
        self.assertEqual(R_SCORED, self.s.getBestMatchingRule().rule_id)
        # Add some more time and frequency should no longer match
        self.db.setTime(11)
        self.assertEqual(None, self.s.getBestMatchingRule())

    def testFactPattern(self):
        """testFactPattern: check for existence of facts matching a pattern"""
        self.s.addRule(serge.blocks.textgenerator.Rule(
            R_SCORED, [
                self.s.FactOccurred(F_ALL_SCORE),
            ]
        ))
        # Initially should be no match
        self.assertEqual(None, self.s.getBestMatchingRule())
        # Still should be no match if wrong rule
        self.db.addFact(F_TEST)
        self.assertEqual(None, self.s.getBestMatchingRule())
        #
        self.db.addFact(F_SCORE)
        self.assertEqual(R_SCORED, self.s.getBestMatchingRule().rule_id)

    def testCombinationOfRules(self):
        """testCombinationOfRules: combine rules to form an add"""
        self.s.addRule(serge.blocks.textgenerator.Rule(
            R_SCORED, [
                self.s.FactOccurred(F_SCORE),
                self.s.FactOccurred(F_BIG_SCORE),
            ]
        ))
        # Initially should be no match
        self.assertEqual(None, self.s.getBestMatchingRule())
        # Still should be no match if wrong rule
        self.db.addFact(F_TEST)
        self.assertEqual(None, self.s.getBestMatchingRule())
        # No match for one part
        self.db.addFact(F_SCORE)
        self.assertEqual(None, self.s.getBestMatchingRule())
        # Match when second fact comes in
        self.db.addFact(F_BIG_SCORE)
        self.assertEqual(R_SCORED, self.s.getBestMatchingRule().rule_id)

    def testRulePriority(self):
        """testRulePriority: can set priority for rules"""
        # Default behaviour is that first rule fires first
        self.s.addRule(serge.blocks.textgenerator.Rule(
            R_SCORED_1, [
                self.s.FactOccurred(F_SCORE),
            ],
        ))
        self.s.addRule(serge.blocks.textgenerator.Rule(
            R_SCORED_2, [
                self.s.FactOccurred(F_SCORE),
            ],
        ))
        self.db.addFact(F_SCORE)
        self.assertEqual(R_SCORED_1, self.s.getBestMatchingRule().rule_id)
        #
        # Use priorities to switch
        self.s = serge.blocks.textgenerator.RuleSystem(self.db)
        self.s.addRule(serge.blocks.textgenerator.Rule(
            R_SCORED_1, [
                self.s.FactOccurred(F_SCORE),
            ],
        ), priority=1)
        self.s.addRule(serge.blocks.textgenerator.Rule(
            R_SCORED_2, [
                self.s.FactOccurred(F_SCORE),
            ],
        ), priority=10)
        # Match when second fact comes in
        self.assertEqual(R_SCORED_2, self.s.getBestMatchingRule().rule_id)

    def testFactNotOccurred(self):
        """testFactNotOccurred: check for fact not occurring"""
        self.s.addRule(serge.blocks.textgenerator.Rule(
            R_SCORED, [
                self.s.FactNotOccurred(F_SCORE),
            ]
        ))
        # Initially should no match
        self.assertEqual(R_SCORED, self.s.getBestMatchingRule().rule_id)
        # Still still match if wrong rule
        self.db.addFact(F_BIG_SCORE)
        self.assertEqual(R_SCORED, self.s.getBestMatchingRule().rule_id)
        # When fact occurs then should stop matching
        self.db.addFact(F_SCORE)
        self.assertEqual(None, self.s.getBestMatchingRule())

    def testCanHaveRuleThatTakesTimeToRecharge(self):
        """testCanHaveRuleThatTakesTimeToRecharge: should be able to have rule that must recharge after firing"""
        self.s.addRule(serge.blocks.textgenerator.RechargingRule(
            R_SCORED, [
                self.s.FactOccurred(F_SCORE),
            ],
            10
        ))
        # Initially should be no match
        self.assertEqual(None, self.s.getBestMatchingRule())
        # Still should be no match if wrong rule
        self.db.addFact(F_BIG_SCORE)
        self.assertEqual(None, self.s.getBestMatchingRule())
        # Should match now
        self.db.addFact(F_SCORE)
        the_rule = self.s.getBestMatchingRule()
        self.assertEqual(R_SCORED, the_rule.rule_id)
        the_rule.fireRule(self.db)
        # Calling again should not match - needs to recharge
        self.assertEqual(None, self.s.getBestMatchingRule())
        self.db.setTime(9)
        self.assertEqual(None, self.s.getBestMatchingRule())
        # After enough time should be charged
        self.db.setTime(10)
        self.assertEqual(R_SCORED, self.s.getBestMatchingRule().rule_id)


if __name__ == '__main__':
    unittest.main()