"""Tests for the Tiled interface"""

import unittest
import os
import pygame

from helper import *

import serge.visual
import serge.registry
import serge.blocks.tiled


class TestTileMap(unittest.TestCase):
    """Tests for the TileMap"""

    def setUp(self):
        """Set up the tests"""
        serge.visual.Sprites.clearItems()
        serge.blocks.tiled.TileMap.resetLayerTypes()
        self.m = serge.blocks.tiled.TileMap()
        self.t1 = [
            [1,1,None],
            [1,1,1],
            [1,1,1],
            [1,1,1],
        ]
                
    def tearDown(self):
        """Tear down the tests"""     

    def testCanAddLayer(self):
        """testCanAddLayer: should be able to add a layer"""
        self.m.addLayer(serge.blocks.tiled.Layer(self.m, 'one', 'visual', tiles=self.t1))
                
    def testCanSetTiles(self):
        """testCanSetTiles: should be able to set tiles"""
        l = self.m.addLayer(serge.blocks.tiled.Layer(self.m, 'one', 'visual', 3, 4))
        for row in range(4):
            for col in range(3):
                l.tiles[row][col] = self.t1[row][col]
        #
        for row in range(4):
            for col in range(3):
                self.assertEqual(l.tiles[row][col], self.t1[row][col])
                        
    def testCanGetLayers(self):
        """testCanGetLayers: should be able to get layers"""
        self.m.addLayer(serge.blocks.tiled.Layer(self.m, 'one', 'visual', tiles=self.t1))
        self.m.addLayer(serge.blocks.tiled.Layer(self.m, 'two', 'movement', 4, 4))
        l = self.m.getLayers()
        self.assertEqual((3,4), l[0].getSize())
        self.assertEqual((4,4), l[1].getSize())
        
    def testCanGetLayersByType(self):
        """testCanGetLayersByType: should be able to get layers by type"""
        self.m.addLayer(serge.blocks.tiled.Layer(self.m, 'one', 'visual', tiles=self.t1))
        self.m.addLayer(serge.blocks.tiled.Layer(self.m, 'two', 'movement', 4, 4))
        l = self.m.getLayerByType('visual')
        self.assertEqual((3,4), l.getSize())
        l = self.m.getLayerByType('movement')
        self.assertEqual((4,4), l.getSize())
        
    def testCanFindTileSetOn(self):
        """testCanFindTileSetOn: should be able to find which layers a tile is set on"""
        self.m.addLayer(serge.blocks.tiled.Layer(self.m, 'one', 'visual', tiles=self.t1))
        self.m.addLayer(serge.blocks.tiled.Layer(self.m, 'two', 'movement', 4, 4))
        self.assertEqual(['visual'], [l.layer_type for l in self.m.getLayersForTile((1,1))])
        self.assertEqual([], [l.layer_type for l in self.m.getLayersForTile((2,0))])
        
    def testCanFindId(self):
        """testCanFindId: should be able to find the Id at a point"""
        self.m.addLayer(serge.blocks.tiled.Layer(self.m, 'one', 'visual', tiles=self.t1))
        self.assertEqual(1, self.m.getLayer('one').tiles[0][0])
                
    def testCanGetSize(self):
        """testCanGetSize: can try to get size"""
        self.m.addLayer(serge.blocks.tiled.Layer(self.m, 'one', 'visual', tiles=self.t1))
        self.assertEqual((3,4), self.m.getSize())
                
        

class TestTiled(unittest.TestCase):
    """Tests for the Tiled"""

    def setUp(self):
        """Set up the tests"""
        serge.visual.Sprites.clearItems()
        serge.blocks.tiled.Tiled.resetLayerTypes()
        
    def tearDown(self):
        """Tear down the tests"""     

    def testCanGetSprites(self):
        """testCanGetSprites: should be able to register sprites from a file"""
        t = serge.blocks.tiled.Tiled(f('world-4.tmx'))
        self.assertEqual(16, serge.visual.Sprites.getItem('tileset-2-1').width)
        self.assertEqual(16, serge.visual.Sprites.getItem('tileset-2-12').width)
        self.assertRaises(serge.registry.UnknownItem, serge.visual.Sprites.getItem, 'tileset-2-0')
        self.assertRaises(serge.registry.UnknownItem, serge.visual.Sprites.getItem, 'tileset-2-13')

    def testCanGetSpriteId(self):
        """testCanGetSpriteId: should be able to get a sprite id"""
        t = serge.blocks.tiled.Tiled(f('world-4.tmx'))
        l1, l2 = t.getLayers()
        name = l1.getSpriteFor((0, 0)).name
        idx = l1.tiles[0][0]
        self.assertEqual(idx, t.getSpriteId(name))

    def testFailMissingFile(self):
        """testFailMissingFile: should fail when file is missing"""
        self.assertRaises(serge.blocks.tiled.BadTiledFile, 
            serge.blocks.tiled.Tiled, f('world-4.tmx.notthere'))
    
    def testMultipleFilesWithSameTilesetIsOk(self):
        """testMultipleFilesWithSameTilesetIsOk: should be able to register with multiple versions of tileset"""
        t = serge.blocks.tiled.Tiled(f('world-4.tmx'))
        t = serge.blocks.tiled.Tiled(f('world-4a.tmx'))

    def testMultipleTilesetsCanGetRightSprite(self):
        """testMultipleTilesetsCanGetRightSprite: should get the right sprite when multiple tilesets"""
        t = serge.blocks.tiled.Tiled(f('world-4.tmx'))
        l1, l2 = t.getLayers()
        self.assertEqual('tileset-2-2', l1.getSpriteFor((0, 0)).name)
        self.assertEqual('tileset-3-1', l2.getSpriteFor((2, 2)).name)

    def testCanSetSpriteFor(self):
        """testCanSetSpriteFor: should be able to set a sprite at a certain location"""
        t = serge.blocks.tiled.Tiled(f('world-4.tmx'))
        l1, l2 = t.getLayers()
        s = serge.visual.Sprites.getItem('tileset-3-1')
        self.assertEqual('tileset-2-2', l1.getSpriteFor((0, 0)).name)
        l1.setSpriteFor((0, 0), s)
        self.assertEqual('tileset-3-1', l2.getSpriteFor((0, 0)).name)

    def testCanCheckLayerType(self):
        """testCanCheckLayerType: should be able to check layer type"""
        t = serge.blocks.tiled.Tiled(f('world-4.tmx'))
        l1, l2 = t.getLayers()
        self.assertEqual('visual', l1.type)
        self.assertEqual('movement', l2.type)
        self.assertEqual(l1, t.getLayerByType('visual'))
        self.assertEqual(l2, t.getLayerByType('movement'))
        self.assertRaises(serge.blocks.tiled.NotFound, t.getLayerByType, 'thing')

    def testCanCheckLayerTypes(self):
        """testCanCheckLayerTypes: should be able to check layer types"""
        t = serge.blocks.tiled.Tiled(f('world-4.tmx'))
        l1, l2 = t.getLayers()
        self.assertEqual('visual', l1.type)
        self.assertEqual('movement', l2.type)
        self.assertEqual([l1], t.getLayersByType('visual'))
        self.assertEqual([l2], t.getLayersByType('movement'))
        self.assertEqual([], t.getLayersByType('thing'))
        
    def testFailBadLayerType(self):
        """testFailBadLayerType: should fail with a bad or missing layer type"""
        self.assertRaises(serge.blocks.tiled.BadLayer, 
            serge.blocks.tiled.Tiled, f('world-4-badtype.tmx'))
        self.assertRaises(serge.blocks.tiled.BadLayer, 
            serge.blocks.tiled.Tiled, f('world-4-missingtype.tmx'))

    def testCanAddLayerType(self):
        """testCanAddLayerType: should be able to add a layer type"""
        serge.blocks.tiled.Tiled.addLayerTypes(['badtype'])
        t = serge.blocks.tiled.Tiled(f('world-4-badtype.tmx'))
        
    def testCanGetTileLayer(self):
        """testCanGetTileLayer: should be able to get a tile layer"""
        t = serge.blocks.tiled.Tiled(f('world-4.tmx'))
        #
        self.assertEqual(2, len(t.getLayers()))
        l1, l2 = t.getLayers()
        #
        self.assertEqual('Tile Layer 1', l1.name)
        self.assertEqual('Tile Layer 2', l2.name)
        self.assertEqual((64, 48), l1.getSize())
        self.assertEqual((64, 48), l2.getSize())
        self.assertEqual(2, l1.tiles[0][0])
        self.assertEqual(13, l2.tiles[0][0])
        #
        # Some properties
        self.assertEqual(123, l1.test1)
        self.assertEqual(123, l1.properties['test1'])
        self.assertEqual(456, l2.test1)
        self.assertEqual('hello', l2.test2)
        self.assertRaises(AttributeError, getattr, l1, 'test2')

    def testFailObjectLayerNotFound(self):
        """testFailObjectLayerNotFound: should fail if asking for object layer with name not found"""
        t = serge.blocks.tiled.Tiled(f('level-1-objects.tmx'))
        self.assertRaises(serge.blocks.tiled.NotFound, t.getObjectLayer, 'not-there')

    def testCanGetObjectLayerNoSize(self):
        """testCanGetObjectLayerNoSize: should be able to get an object layer with no size"""
        t = serge.blocks.tiled.Tiled(f('level-1-objects.tmx'))
        #
        # Layer size should be 0, 0
        l = t.getObjectLayer('Background')
        self.assertEqual(-1, l.getSize()[0])
        self.assertEqual(-1, l.getSize()[1])

    def testCanGetObjectLayerWithSprites(self):
        """testCanGetObjectLayerWithSprites: should be able to get an object layer that has sprites on it"""
        t = serge.blocks.tiled.Tiled(f('level-1-objects.tmx'))
        l = t.getObjectLayer('Background')
        objects = l.getObjects()
        #
        self.assertEqual(32, len(objects))
        self.assertEqual('sprite', objects[0].object_type)
        self.assertEqual(379.324 + 50, objects[0].x)
        self.assertEqual(545.477 - 50, objects[0].y) # Note that tiled reports y-coord as the bottom of the sprite
        self.assertEqual(2, objects[0].sprite_id)
        self.assertEqual('Bricks-2', objects[0].sprite_name)

    def testCanGetRotationForSpritesInObjectLayer(self):
        """testCanGetRotationForSpritesInObjectLayer: should be able to get rotation sprite layers"""
        t = serge.blocks.tiled.Tiled(f('level-1-objects.tmx'))
        l = t.getObjectLayer('Background')
        objects = l.getObjects()
        self.assertEqual(0, objects[0].rotation)
        self.assertEqual(-47.4366, objects[-1].rotation)

    def testCanGetObjectLayerWithPolyLine(self):
        """testCanGetObjectLayerWithPolyLine: should be able to get an object layer with poly lines on it"""
        t = serge.blocks.tiled.Tiled(f('level-1-objects.tmx'))
        l = t.getObjectLayer('Geometry')
        objects = l.getObjects()
        sx, sy = 9.45052, 517.078
        self.assertEqual(1, len(objects))
        self.assertEqual('polyline', objects[0].object_type)
        self.assertEqual(17, len(objects[0].points))
        self.assertEqual([sx, sy], objects[0].points[0])
        self.assertEqual([sx + 194.411, sy + 2.70015], objects[0].points[1])
        self.assertEqual([sx + 791.144, sy - 225.462], objects[0].points[-1])

    def testCanGetImageLayer(self):
        """testCanGetImageLayer: should be able to get an image layer"""
        t = serge.blocks.tiled.Tiled(f('level-2-objects.tmx'))
        l = t.getObjectLayer('Backdrop')
        objects = l.getObjects()
        self.assertEqual(1, len(objects))
        obj = objects[0]
        self.assertEqual(1 + 48 / 2, obj.x)
        self.assertEqual(-1 + 64 / 2, obj.y)
        self.assertEqual('/Users/paul/Workspace/git/perpetualpyramid/serge/test/files/tileset-2.png', obj.filename)
        #
        # Check the surface
        surface = obj.getSurface()
        self.assertEqual((0, 0, 255), surface.get_at((47, 2)))

    def testCanGetObjectLayers(self):
        """testCanGetObjectLayers: should be able to get an object layer"""
        t = serge.blocks.tiled.Tiled(f('world-4.tmx'))
        #
        layers = t.getObjectLayers()
        self.assertEqual(3, len(layers))
        l1, l2, l3 = layers
        #
        # Properties
        self.assertEqual(True, l1.AllMissions)
        self.assertEqual(True, l2.Mission1a)
        self.assertEqual(True, l3.Mission1b)
        #
        # Objects
        self.assertEqual('All Missions', l1.name)
        self.assertEqual(4, len(l1.getObjects()))
        td = l1.getObject('TestDoor1')
        self.assertEqual('Door', td.object_type)
        self.assertEqual('world-5', td.leads_to)
        self.assertEqual('shop-west', td.location)
        self.assertEqual(round(41.125*16)+td.width/2, td.x)
        self.assertEqual(round(25.250*16)+td.height/2, td.y)
        self.assertEqual((round(0.688*16), round(3.562*16)), (td.width, td.height))
        #
        self.assertEqual('Mission1a', l2.name)
        tb = l2.getObject('Bob')
        self.assertEqual((0, 0), (tb.width, tb.height))

    def testCanDoSpritesOnObjectLayers(self):
        """testCanDoSpritesOnObjectLayers: should be able to do sprites on object layers"""
        t = serge.blocks.tiled.Tiled(f('world-5.tmx'))
        #
        layers = t.getLayersByType('adhoc-visual')
        self.assertEqual(1, len(layers))
        objects = layers[0].getObjects()
        self.assertEqual(6, len(objects))
        self.assertEqual('tileset-2-6', objects[0].sprite_name)
        
    def testCanDoPolygonObjects(self):
        """testCanDoPolygonObjects: should be able to do polygon objects"""
        t = serge.blocks.tiled.Tiled(f('world-6.tmx'))
        #
        layers = t.getObjectLayers()
        self.assertEqual(1, len(layers))
        #
        objects = layers[0].getObjects()
        self.assertEqual(3, len(objects))
        #
        obj1, obj2, obj3 = objects
        self.assertEqual('one', obj1.name)
        self.assertEqual((184, 108), (int(obj1.x), int(obj1.y)))
        self.assertEqual((77, 90), (int(obj1.width), int(obj1.height)))
        #
        self.assertEqual('two', obj2.name)
        self.assertEqual((86, 106), (int(obj2.width), int(obj2.height)))
        #
        self.assertEqual('three', obj3.name)
        self.assertEqual(((547-24)+obj3.width/2, (278-25)+obj3.height/2), (obj3.x, obj3.y))
        self.assertEqual((81--24, 89--25), (obj3.width, obj3.height))

    def testFailGettingMissingObject(self):
        """testFailGettingMissingObject: should fail when getting mission object"""
        t = serge.blocks.tiled.Tiled(f('world-4.tmx'))
        #
        layers = t.getObjectLayers()
        l1, l2, l3 = layers
        self.assertRaises(serge.blocks.tiled.NotFound, l1.getObject, 'TestDoor1NotThere')                
    
    def testCanCheckWhichLayersATileIsSetOn(self):
        """testCanCheckWhichLayersATileIsSetOn: should be able to check which layers has a tile set on"""
        t = serge.blocks.tiled.Tiled(f('world-4.tmx'))
        self.assertEqual(['visual', 'movement'], [l.type for l in t.getLayersForTile((0, 0))])
        self.assertEqual(['visual'], [l.type for l in t.getLayersForTile((5, 4))])
        self.assertEqual(['visual', 'movement'], [l.type for l in t.getLayersForTile((37, 6))])
        
    def testCanCheckLayersExcludingSome(self):
        """testCanCheckLayersExcludingSome: should be able to check which layers are set with exclusions"""
        t = serge.blocks.tiled.Tiled(f('world-4.tmx'))
        self.assertEqual(['movement'], [l.type for l in t.getLayersForTile((0, 0), excluding=['visual'])])
        
    def testCanFindSetOn(self):
        """testCanFindSetOn: should be able to find the tiles that are set on a layer"""
        t = serge.blocks.tiled.Tiled(f('world-4.tmx'))
        #
        l = t.getLayer("Tile Layer 2")
        set_tiles = set(l.getLocationsWithTile())
        unset_tiles = set(l.getLocationsWithoutTile()) 
        #
        self.assertEqual(64*48, len(unset_tiles) + len(set_tiles))
        self.assertEqual(set(), unset_tiles.intersection(set_tiles))
        self.assertTrue(len(unset_tiles) > len(set_tiles))
        self.assertFalse(len(set_tiles) == 0)
        self.assertTrue((0,0) in set_tiles)
        self.assertTrue((35, 10) in set_tiles)
        self.assertFalse((14, 7) in set_tiles)
    
    def testCanFindSpecificTileSetOn(self):
        """testCanFindSpecificTileSetOn: should be able to check locations a tile is set on"""
        t = serge.blocks.tiled.Tiled(f('world-4.tmx'))
        #
        l = t.getLayer("Tile Layer 1")
        set_tiles = set(l.getLocationsWithSpriteName('tileset-2-9'))
        self.assertSetEqual(
            set([(41,25), (41,26), (41,27), (41,28),
                 (50,25), (50,26), (50,27), (50,28)]),
            set_tiles)

    def testCanGetPropertyBagArray(self):
        """testCanGetPropertyBagArray: should be able to get a property bag array"""
        t = serge.blocks.tiled.Tiled(f('world-5.tmx'))
        #
        array = t.getPropertyBagArray(
            sprite_layers=['visual'],
            boolean_layers=['movement', 'visibility'],
            property_layers=['resistance'])
        #
        a00 = array[0][0]
        self.assertEqual(False, a00.visibility)
        self.assertEqual(True, a00.movement)
        self.assertEqual(['tileset-2-2', 'tileset-3-9'], a00.sprites)
        self.assertFalse(hasattr(a00, 'resistance'))
        a12 = array[2][1]
        self.assertEqual(False, a12.visibility)
        self.assertEqual(False, a12.movement)
        self.assertEqual(['tileset-2-2', 'tileset-3-9'], a12.sprites)
        self.assertEqual(10, a12.resistance)
        a14 = array[4][1]
        self.assertEqual(['tileset-2-2'], a14.sprites)
        a54 = array[4][5]
        self.assertEqual(['tileset-2-1'], a54.sprites)
        self.assertEqual(30, a54.resistance)
        
    def testPropertyBagCanAcceptPrototype(self):
        """testPropertyBagCanAcceptPrototype: should be able to pass prototype for Bag"""
        t = serge.blocks.tiled.Tiled(f('world-5.tmx'))
        #
        default = serge.serialize.SerializedBag(resistance=0)
        array = t.getPropertyBagArray(
            sprite_layers=['visual'],
            boolean_layers=['movement', 'visibility'],
            property_layers=['resistance'],
            prototype=default)
        #
        a00 = array[0][0]
        self.assertEqual(0, a00.resistance)
        a12 = array[2][1]
        self.assertEqual(10, a12.resistance)
        a54 = array[4][5]
        self.assertEqual(30, a54.resistance)
        
    def testPropertyBagFailIfMissingLayer(self):
        """testPropertyBagFailIfMissingLayer: should fail on property bag when a layer is missing"""
        t = serge.blocks.tiled.Tiled(f('world-5.tmx'))
        #
        self.assertRaises(serge.blocks.tiled.NotFound, t.getPropertyBagArray,
            sprite_layers=['visual'],
            boolean_layers=['missing-layer-type', 'movement', 'visibility'],
            property_layers=['resistance'])
        
    def testPropertyBagCanHaveOptionalLayers(self):
        """testPropertyBagCanHaveOptionalLayers: should be able to use optional layers on the property bag"""
        t = serge.blocks.tiled.Tiled(f('world-5.tmx'))
        #
        array = t.getPropertyBagArray(
            sprite_layers=['visual'],
            boolean_layers=['missing_layer_type', 'movement', 'visibility'],
            property_layers=['resistance'],
            optional_layers=['missing_layer_type'])
        #
        self.assertFalse(array[0][0].missing_layer_type)

    def testCanUseClassPropertySchemas(self):
        """testCanUseClassPropertySchemas: should be able to set a class property schema"""
        t = serge.blocks.tiled.Tiled(f('level-4-object-schema.tmx'))
        l = t.getObjectLayer('Lights')
        self.assertTrue(l.validateSchemas({
            'directional-light': [
                'angle', 'beam_width', 'distance', 'intensity', 'particle', 'randomize',
            ]
        }))

    def testFailIfSchemaNotMet(self):
        """testFailIfSchemaNotMet: should fail to load an object if schema is not met"""
        t = serge.blocks.tiled.Tiled(f('level-4-object-schema.tmx'))
        l = t.getObjectLayer('Lights')
        self.assertRaises(serge.blocks.tiled.SchemaNotMet, l.validateSchemas, {
            'directional-light': [
                'angle', 'beam_width', 'distance', 'intensity', 'particle', 'randomize', 'NOTTHERE',
            ]
        })

    def testSchemaShouldAllowExtraProperties(self):
        """testSchemaShouldAllowExtraProperties: schema should be able to have extras"""
        t = serge.blocks.tiled.Tiled(f('level-4-object-schema.tmx'))
        l = t.getObjectLayer('Lights')
        self.assertTrue(l.validateSchemas({
            'directional-light': [
                'angle', 'beam_width', 'distance', 'intensity', 'particle', ##'randomize',
            ]
        }))

    def testCanHaveStrictSchema(self):
        """testCanHaveStrictSchema: should be able to have strict schema"""
        t = serge.blocks.tiled.Tiled(f('level-4-object-schema.tmx'))
        l = t.getObjectLayer('Lights')
        self.assertRaises(serge.blocks.tiled.SchemaNotMet, l.validateSchemas, {
            'directional-light': [
                'angle', 'beam_width', 'distance', 'intensity', 'particle', ##'randomize',
            ]
        }, strict_mode=True)

    def testStrictModeChecksAllTypes(self):
        """testStrictModeChecksAllTypes: strict mode should detect objects without a valid type"""
        t = serge.blocks.tiled.Tiled(f('level-4-object-schema.tmx'))
        l = t.getObjectLayer('Missing-type')
        self.assertRaises(serge.blocks.tiled.SchemaNotMet, l.validateSchemas, {
            'directional-light': [
                'angle', 'beam_width', 'distance', 'intensity', 'particle', 'randomize',
            ]
        }, strict_mode=True)

    def testShouldFailIfHasTypeNotInSchema(self):
        """testShouldFailIfHasTypeNotInSchema: if an object has a type but it isn't in the schema then should fail"""
        t = serge.blocks.tiled.Tiled(f('level-4-object-schema.tmx'))
        l = t.getObjectLayer('Lights')
        self.assertRaises(serge.blocks.tiled.SchemaNotMet, l.validateSchemas, {})

    def testCanHaveOptionalParameters(self):
        """testCanHaveOptionalParameters: a schema should be able to have optional parameters"""
        t = serge.blocks.tiled.Tiled(f('level-4-object-schema.tmx'))
        l = t.getObjectLayer('Optional-type')
        self.assertTrue(l.validateSchemas({
            'directional-light': [
                'angle', 'beam_width', 'distance', 'intensity', 'particle', '?randomize',
            ]
        }, strict_mode=True))


if __name__ == '__main__':
    unittest.main()
