"""Test of the timer blocks"""

import unittest
import serge.blocks.timers


class TimerTester(unittest.TestCase):
    """A base tester for timers"""

    def setUp(self):
        """Set up the tests"""
        self._fired = False
        self._times = 0

    def tearDown(self):
        """Tear down the tests"""

    def _timerFired(self):
        """A timer fired"""
        self._fired = True
        self._times += 1


class TestRandomTimer(TimerTester):
    """Tests for the RandomTimer"""

    def testCanCreateTimer(self):
        """testCanCreateTimer: should be able to create a timer"""
        _ = serge.blocks.timers.Timer('t', 't', 10, 20, self._timerFired)

    def testTimerFiresInRange(self):
        """testTimerFiresInRange: timer should fire within its range"""
        tm = serge.blocks.timers.Timer('t', 't', 10, 20, self._timerFired)
        #
        self.assertFalse(self._fired)
        self.assertEqual(0, self._times)
        #
        tm.updateActor(9999, None)
        self.assertFalse(self._fired)
        self.assertEqual(0, self._times)
        #
        tm.updateActor(10001, None)
        self.assertTrue(self._fired)
        self.assertEqual(1, self._times)

    def testFiresMultipleTimes(self):
        """testFiresMultipleTimes: should fire multiple times"""
        tm = serge.blocks.timers.Timer('t', 't', 10, callback=self._timerFired)
        tm.updateActor(30000, None)
        #
        self.assertTrue(self._fired)
        self.assertEqual(3, self._times)

    def testCanCreateInStoppedMode(self):
        """testCanCreateInStoppedMode: should be able to create with timer stopped"""
        tm = serge.blocks.timers.Timer('t', 't', 10, callback=self._timerFired, started=False)
        tm.updateActor(30000, None)
        #
        self.assertFalse(self._fired)

    def testCanDoOneShot(self):
        """testCanDoOneShot: should be able to do one shot"""
        tm = serge.blocks.timers.Timer('t', 't', 10, callback=self._timerFired, one_shot=True)
        tm.updateActor(30000, None)
        #
        self.assertTrue(self._fired)
        self.assertEqual(1, self._times)

    def testCanPauseTimer(self):
        """testCanPauseTimer: should be able to pause a timer"""
        tm = serge.blocks.timers.Timer('t', 't', 10, callback=self._timerFired)
        #
        tm.updateActor(5000, None)
        tm.stopTimer()
        tm.updateActor(5000, None)
        self.assertFalse(self._fired)

    def testCanResumeTimer(self):
        """testCanResumeTimer: should be able to resume timer"""
        tm = serge.blocks.timers.Timer('t', 't', 10, callback=self._timerFired)
        #
        tm.updateActor(5000, None)
        tm.stopTimer()
        tm.startTimer()
        tm.updateActor(5000, None)
        self.assertTrue(self._fired)

    def testFailHighLessThanLow(self):
        """testFailHighLessThanLow: should throw an error if high time less than low time"""
        self.assertRaises(
            serge.blocks.timers.BadRange,
            serge.blocks.timers.Timer, 't', 't', 10, 9, self._timerFired
        )

    def testShouldBeAbleToHaveLeastAndMostSame(self):
        """testShouldBeAbleToHaveLeastAndMostSame: should be able to have most and least the same"""
        tm = serge.blocks.timers.Timer('t', 't', 10, 10, callback=self._timerFired)
        tm.updateActor(10000, None)
        self.assertTrue(self._fired)
        self.assertEqual(1, self._times)

    def testCanResetTimer(self):
        """testCanResetTimer: should be able to reset timer"""
        tm = serge.blocks.timers.Timer('t', 't', 10, callback=self._timerFired)
        #
        tm.updateActor(5000, None)
        tm.resetTimer()
        tm.updateActor(5000, None)
        self.assertFalse(self._fired)
        #
        # Should still be running
        tm.updateActor(5000, None)
        self.assertTrue(self._fired)

    def testCanResetAndStop(self):
        """testCanResetAndStop: should be able to reset and stop"""
        tm = serge.blocks.timers.Timer('t', 't', 10, callback=self._timerFired)
        #
        tm.updateActor(5000, None)
        tm.resetAndStopTimer()
        tm.updateActor(5000, None)
        self.assertFalse(self._fired)
        #
        # Should not still be running
        tm.updateActor(5000, None)
        self.assertFalse(self._fired)

    def testCanResetAndStart(self):
        """testCanResetAndStart: should be able to reset and start the timer"""
        tm = serge.blocks.timers.Timer('t', 't', 10, callback=self._timerFired)
        #
        tm.updateActor(5000, None)
        tm.stopTimer()
        tm.updateActor(5000, None)
        self.assertFalse(self._fired)
        #
        # Reset and continue
        tm.resetAndStartTimer()
        tm.updateActor(5000, None)
        self.assertFalse(self._fired)
        tm.updateActor(5000, None)
        self.assertTrue(self._fired)

    def testCanStopTimer(self):
        """testCanStopTimer: should be able to stop timer"""
        tm = serge.blocks.timers.Timer('t', 't', 10, callback=self._timerFired)
        #
        tm.updateActor(5000, None)
        tm.stopTimer()
        tm.updateActor(5000, None)
        self.assertFalse(self._fired)
        #
        # Should not still be running
        tm.updateActor(50000, None)
        self.assertFalse(self._fired)

    def testCanRestartTimer(self):
        """testCanRestartTimer: should be able to restart timer"""
        tm = serge.blocks.timers.Timer('t', 't', 10, callback=self._timerFired)
        #
        tm.updateActor(5000, None)
        tm.stopTimer()
        tm.updateActor(5000, None)
        self.assertFalse(self._fired)
        #
        # Should be able to start where we left off
        tm.startTimer()
        tm.updateActor(5000, None)
        self.assertTrue(self._fired)

    def testDoubleStopDoesNothing(self):
        """testDoubleStopDoesNothing: should be able to stop a stopped timer"""
        tm = serge.blocks.timers.Timer('t', 't', 10, callback=self._timerFired)
        #
        tm.updateActor(5000, None)
        tm.stopTimer()
        tm.updateActor(5000, None)
        self.assertFalse(self._fired)
        #
        # Should be able to start where we left off
        tm.startTimer()
        tm.updateActor(5000, None)
        self.assertTrue(self._fired)

    def testFrequencyShouldBeRightIfGoOver(self):
        """testFrequencyShouldBeRightIfGoOver: timing should respect actual frequency"""
        tm = serge.blocks.timers.Timer('t', 't', 10, callback=self._timerFired)
        #
        tm.updateActor(15000, None)
        self.assertTrue(self._fired)
        self.assertEqual(1, self._times)
        #
        tm.updateActor(5000, None)
        self.assertEqual(2, self._times)

    def testCanCheckIfRunning(self):
        """testCanCheckIfRunning: should be able to check if running"""
        tm = serge.blocks.timers.Timer('t', 't', 10)
        self.assertTrue(tm.isRunning())
        tm.stopTimer()
        #
        tm = serge.blocks.timers.Timer('t', 't', 10, started=False)
        self.assertFalse(tm.isRunning())
        tm.startTimer()
        self.assertTrue(tm.isRunning())


class TestSequenceIntervalTimer(TimerTester):
    """Tests for the SequenceTimer"""

    def testBlankSequenceStops(self):
        """testBlankSequenceStops: when given a blank sequence should stop"""
        tm = serge.blocks.timers.SequenceIntervalTimer('t', 't', [], callback=self._timerFired)
        self.assertFalse(tm.isRunning())
        tm.updateActor(1, None)
        self.assertFalse(tm.isRunning())
        self.assertEqual(0, self._times)

    def testSequenceWorks(self):
        """testSequenceWorks: sequence should behave as expected"""
        tm = serge.blocks.timers.SequenceIntervalTimer('t', 't', [0.1, 0.2, 0.3], callback=self._timerFired)
        tm.updateActor(99, None)
        self.assertEqual(0, self._times)
        # Up to 100
        tm.updateActor(1, None)
        self.assertEqual(1, self._times)
        # Up to 200
        tm.updateActor(100, None)
        self.assertEqual(1, self._times)
        # Up to 300
        tm.updateActor(100, None)
        self.assertEqual(2, self._times)
        # Up to 400
        tm.updateActor(200, None)
        self.assertEqual(2, self._times)
        # Up to 600
        tm.updateActor(200, None)
        self.assertEqual(3, self._times)

    def testStopsWhenReachesEndOfSequence(self):
        """testStopsWhenReachesEndOfSequence: should stop when reaching the end of sequence"""
        tm = serge.blocks.timers.SequenceIntervalTimer('t', 't', [0.1, 0.2, 0.3], callback=self._timerFired)
        tm.updateActor(800, None)
        self.assertEqual(3, self._times)
        self.assertFalse(tm.isRunning())

    def testCanRepeatSequence(self):
        """testCanRepeatSequence: should be able to repeat a sequence"""
        tm = serge.blocks.timers.SequenceIntervalTimer('t', 't', [0.1, 0.2, 0.3], repeat=True,
                                                       callback=self._timerFired)
        tm.updateActor(650, None)
        self.assertEqual(3, self._times)
        self.assertTrue(tm.isRunning())
        #
        tm.updateActor(50, None)
        self.assertEqual(4, self._times)
        self.assertTrue(tm.isRunning())

    def testFailIfNegativeTime(self):
        """testFailIfNegativeTime: should fail if some intervals are negative"""
        self.assertRaises(
            serge.blocks.timers.BadSequence,
            serge.blocks.timers.SequenceIntervalTimer, 't', 't', [100, -200, 300]
        )

    def testCanRestartSequence(self):
        """testCanRestartSequence: should be able to restart the sequence"""
        tm = serge.blocks.timers.SequenceIntervalTimer('t', 't', [0.1, 0.2, 0.3], callback=self._timerFired)
        tm.updateActor(301, None)
        self.assertEqual(2, self._times)
        self.assertTrue(tm.isRunning())
        #
        tm.resetAndStartTimer()
        tm.updateActor(100, None)
        self.assertEqual(3, self._times)


class TestSequenceTimer(TimerTester):
    """Tests for the SequenceTimer"""

    def testSequenceWorks(self):
        """testSequenceWorks: should be able to give it a sequence"""
        tm = serge.blocks.timers.SequenceTimer('t', 't', [0.1, 0.3, 0.6], callback=self._timerFired)
        tm.updateActor(99, None)
        self.assertEqual(0, self._times)
        # Up to 100
        tm.updateActor(1, None)
        self.assertEqual(1, self._times)
        # Up to 200
        tm.updateActor(100, None)
        self.assertEqual(1, self._times)
        # Up to 300
        tm.updateActor(100, None)
        self.assertEqual(2, self._times)
        # Up to 400
        tm.updateActor(200, None)
        self.assertEqual(2, self._times)
        # Up to 600
        tm.updateActor(200, None)
        self.assertEqual(3, self._times)

    def testFailIfSequenceNotMonotonic(self):
        """testFailIfSequenceNotMonotonic: should fail if sequence is not monotonic"""
        self.assertRaises(
            serge.blocks.timers.BadSequence,
            serge.blocks.timers.SequenceTimer, 't', 't', [100, 200, 200, 300]
        )

    def testEmptySequenceShouldStop(self):
        """testEmptySequenceShouldStop: should stop immediately on empty sequence"""
        tm = serge.blocks.timers.SequenceTimer('t', 't', [], callback=self._timerFired)
        self.assertFalse(tm.isRunning())
        tm.updateActor(1, None)
        self.assertFalse(tm.isRunning())
        self.assertEqual(0, self._times)


if __name__ == '__main__':
    unittest.main()