"""Tests for the utilities that speed up common operations"""

import unittest
import collections

import pygame
import os

pygame.init()

import serge.blocks.utils

import serge.render
import serge.actor
import serge.zone
import serge.world
import serge.engine
import serge.visual


class TestUtils(unittest.TestCase):
    """Tests for the Utils"""

    def setUp(self):
        """Set up the tests"""
        serge.visual.Register.clearItems()
        
    def tearDown(self):
        """Tear down the tests"""

    def testCanSetLayersForEngine(self):
        """testCanSetLayersForEngine: should be able to quickly set layers"""
        e = serge.engine.Engine()
        serge.blocks.utils.createLayersForEngine(e, ['one','two','three'])
        r = e.getRenderer()
        #
        # Should be the three layers that we created
        l1 = r.getLayer('one')
        l2 = r.getLayer('two')
        l3 = r.getLayer('three')
        #
        # Should have the right ordering
        self.assertEqual(0, l1.order)
        self.assertEqual(1, l2.order)
        self.assertEqual(2, l3.order)
        #
        # And be active
        self.assertEqual(True, l1.active)
        self.assertEqual(True, l2.active)
        self.assertEqual(True, l3.active)
        #
        # Adding another should add one more in sequence
        serge.blocks.utils.createLayersForEngine(e, ['four'])
        l4 = r.getLayer('four')
        self.assertEqual(3, l4.order)
        self.assertEqual(True, l4.active)

    def testCanSetVirtualLayersForEngine(self):
        """testCanSetVirtualLayersForEngine: should be able to quickly set virtual layers"""
        e = serge.engine.Engine()
        serge.blocks.utils.createVirtualLayersForEngine(e, ['one', 'two', 'three'])
        r = e.getRenderer()
        #
        # Should be the three layers that we created
        l1 = r.getLayer('one')
        l2 = r.getLayer('two')
        l3 = r.getLayer('three')
        #
        # Should be virtual
        self.assertEqual(r.getSurface(), l1.getSurface())
        self.assertEqual(r.getSurface(), l2.getSurface())
        self.assertEqual(r.getSurface(), l3.getSurface())
        #
        # Should have the right ordering
        self.assertEqual(0, l1.order)
        self.assertEqual(1, l2.order)
        self.assertEqual(2, l3.order)
        #
        # And be active
        self.assertEqual(True, l1.active)
        self.assertEqual(True, l2.active)
        self.assertEqual(True, l3.active)
        #
        # Adding another should add one more in sequence
        serge.blocks.utils.createLayersForEngine(e, ['four'])
        l4 = r.getLayer('four')
        self.assertEqual(3, l4.order)
        self.assertEqual(True, l4.active)

    def testCanAddWorldsWithZones(self):
        """testCanAddWorldsWithZones: should be able to quickly add worlds with big zones"""
        e = serge.engine.Engine()
        serge.blocks.utils.createWorldsForEngine(e, ['one', 'two', 'three'])
        #
        # Should be three worlds created
        w1 = e.getWorld('one')
        w2 = e.getWorld('two')
        w3 = e.getWorld('three')
        #
        # Each should have one zone
        self.assertEqual(1, len(w1.zones))
        self.assertEqual(1, len(w2.zones))
        self.assertEqual(1, len(w3.zones))
        #
        # Zones should be active and large
        for w in (w1, w2, w3):
            z = list(w.zones)[0]
            self.assertEqual(True, z.active)   
            self.assert_(z.getSpatial()[0] < -1000)
            self.assert_(z.getSpatial()[1] < -1000)
            self.assert_(z.width > 2000)
            self.assert_(z.height > 2000)

    def testCanAddSpriteActorToWorld(self):
        """testCanAddSpriteActorToWorld: should be able to quickly add an actor for a sprite to the world"""
        e = serge.engine.Engine()
        serge.blocks.utils.createWorldsForEngine(e, ['one'])
        w = e.getWorld('one')
        serge.visual.Register.setPath(os.path.join(os.path.abspath(os.curdir), 'test', 'images'))
        serge.visual.Register.registerItem('b', 'bluerect.png')        
        #
        # Add it and it should be returned
        a = serge.blocks.utils.addSpriteActorToWorld(w, 'tag', 'name', 'b', 'layer', (100, 200))
        self.assertEqual(a, w.findActorByName('name'))
        #
        # Should be positioned where we wanted it
        self.assertEqual(100, a.x)
        self.assertEqual(200, a.y)
        #
        # Should be on the right layer and have the right metadata
        self.assertEqual('layer', a.getLayerName())
        self.assertEqual('tag', a.tag)
        self.assertEqual('name', a.name)
        #
        # If don't specify a position then it should be centered
        b = serge.blocks.utils.addSpriteActorToWorld(w, 'tag', 'b', 'b', 'layer')
        self.assertEqual(320, b.x)
        self.assertEqual(240, b.y)

    def testCanAddVisualActorToWorld(self):
        """testCanAddVisualActorToWorld: should be able to quickly add an actor with a visual to the world"""
        e = serge.engine.Engine()
        serge.blocks.utils.createWorldsForEngine(e, ['one'])
        w = e.getWorld('one')
        serge.visual.Register.setPath(os.path.join(os.path.abspath(os.curdir), 'test', 'images'))
        serge.visual.Register.registerItem('b', 'bluerect.png')        
        #
        # Add it and it should be returned
        visual = serge.visual.Sprite()
        a = serge.blocks.utils.addVisualActorToWorld(w, 'tag', 'name', visual, 'layer', (100, 200))
        self.assertEqual(a, w.findActorByName('name'))
        #
        # Should be positioned where we wanted it
        self.assertEqual(100, a.x)
        self.assertEqual(200, a.y)
        #
        # Should be on the right layer and have the right metadata
        self.assertEqual('layer', a.getLayerName())
        self.assertEqual('tag', a.tag)
        self.assertEqual('name', a.name)
        #
        # If don't specify a position then it should be centered
        b = serge.blocks.utils.addSpriteActorToWorld(w, 'tag', 'b', 'b', 'layer')
        self.assertEqual(320, b.x)
        self.assertEqual(240, b.y)
        
    def testCanAddActorToWorld(self):
        """testCanAddActorToWorld: should be able to quickly add an actor to the world"""
        e = serge.engine.Engine()
        serge.blocks.utils.createWorldsForEngine(e, ['one'])
        w = e.getWorld('one')
        serge.visual.Register.setPath(os.path.join(os.path.abspath(os.curdir), 'test', 'images'))
        serge.visual.Register.registerItem('b', 'bluerect.png')        
        #
        # Add it and it should be returned
        visual = serge.visual.Sprite()
        actor1 = serge.actor.Actor('tag', 'name')
        a = serge.blocks.utils.addActorToWorld(w, actor1, 'b', 'layer', (100, 200))
        self.assertEqual(a, w.findActorByName('name'))
        #
        # Should be positioned where we wanted it
        self.assertEqual(100, a.x)
        self.assertEqual(200, a.y)
        #
        # Should be on the right layer and have the right metadata
        self.assertEqual('layer', a.getLayerName())
        self.assertEqual('tag', a.tag)
        self.assertEqual('name', a.name)
        #
        # If don't specify a position then it should be centered
        actor2 = serge.actor.Actor('tag', 'name')
        b = serge.blocks.utils.addActorToWorld(w, actor2, 'b', 'layer')
        self.assertEqual(320, b.x)
        self.assertEqual(240, b.y)
    
    def testRandomFromEqualProbability(self):
        """testRandomFromEqualProbability: should be able to get items from equal random probability"""
        p = {
            'a': 0.1,
            'b': 0.1,
            'c': 0.1,
        }
        chooser = serge.blocks.utils.ProbabilityChooser(p)
        counts = collections.defaultdict(int)
        #
        # Choose lots of times
        for idx in range(3000):
            choice = chooser.choose()
            counts[choice] += 1
        #
        # Results should be almost equal to 1000 each
        self.assertTrue(abs(counts['a'] - 1000) <= 50)
        self.assertTrue(abs(counts['b'] - 1000) <= 50)
        self.assertTrue(abs(counts['c'] - 1000) <= 50)

    def testRandomFromNonEqualProbability(self):
        """testRandomFromNonEqualProbability: should be able to get items from non equal random probability"""
        p = {
            'a': 0.1,
            'b': 0.2,
            'c': 0.3,
        }
        chooser = serge.blocks.utils.ProbabilityChooser(p)
        counts = collections.defaultdict(int)
        #
        # Choose lots of times
        for idx in range(6000):
            choice = chooser.choose()
            counts[choice] += 1
        #
        # Results should be almost equal to 1000 each
        self.assertTrue(abs(counts['a'] - 1000) <= 50)
        self.assertTrue(abs(counts['b'] - 2000) <= 50)
        self.assertTrue(abs(counts['c'] - 3000) <= 50)

    def testCanChooseMultiple(self):
        """testCanChooseMultiple: can choose multiple items randomly"""
        p = {
            'a': 0.1,
            'b': 0.2,
            'c': 0.3,
        }
        chooser = serge.blocks.utils.ProbabilityChooser(p)
        items = chooser.chooseMultiple(100)
        self.assertEqual(100, len(items))
        self.assertTrue('a' in items)
        self.assertTrue('b' in items)
        self.assertTrue('c' in items)

    def testBidirectionalDictionary(self):
        """testBidirectionalDictionary: test of bidirectional dictionary"""
        d = serge.blocks.utils.BidirectionalDict()
        self.assertEqual(0, len(d))
        d['1'] = 1
        self.assertEqual(1, len(d))
        self.assertEqual(1, d['1'])
        self.assertEqual('1', d.getReverse(1))
        #
        d['2'] = 2
        self.assertEqual(2, len(d))
        self.assertEqual(2, d['2'])
        self.assertEqual('2', d.getReverse(2))
        #
        self.assertSetEqual(set(['1', '2']), set(d.keys()))
        self.assertSetEqual(set(['1', '2']), set(list(d)))
        self.assertSetEqual(set([1, 2]), set(d.values()))
        self.assertSetEqual(set([('1', 1), ('2', 2)]), set(d.iteritems()))
        #
        del(d['2'])
        self.assertEqual(1, len(d))
        self.assertRaises(KeyError, d.__getitem__, '2')
        self.assertRaises(KeyError, d.__getitem__, 2)
        #
        d.deleteReverse(1)
        self.assertEqual(0, len(d))
        self.assertRaises(KeyError, d.__getitem__, '1')
        self.assertRaises(KeyError, d.__getitem__, 1)
    

if __name__ == '__main__':
    unittest.main()
