"""Test of the model class"""

import os
import unittest
from game import model


# Files
ENEMY_FILE = os.path.join('..', 'game', 'enemies.csv')
POWER_UP_FILE = os.path.join('..', 'game', 'powerups.csv')

model.LivingThing.timer_randomness = (1.0, 1.0)


class TestModelSetup(unittest.TestCase):
    """Test of the Model class"""

    def setUp(self):
        """Set up the tests"""
        self.e = model.Player()
        self.l = model.Enemy()
        self.s = model.Sidekick()

    def tearDown(self):
        """Tear down the tests"""

    def testCanCreateEngine(self):
        """testCanCreateEngine: should be able to create an engine"""
        # Check properties
        self.assertGreater(self.e.hp, 0)
        self.assertGreater(self.e.click_damage, 0)
        self.assertGreater(self.e.critical_click_probability, 0)
        self.assertGreater(self.e.critical_click_multiplier, 0)
        self.assertEqual(self.e.purse, 0)
        self.assertGreater(self.e.max_purse_size, 0)
        self.assertGreater(self.e.ac, 0)
        self.assertEqual(self.e.auto_click_interval, 0)
        self.assertEqual(self.e.click_aof_damage, 0)

    def testCanCreateEnemy(self):
        """testCanCreateEnemy: should be able to create an enemy"""
        # Check properties
        self.assertGreater(self.l.hp, 0)

    def testCanCreateSideKick(self):
        """testCanCreateSideKick: should be able to create a sidekick"""
        self.assertGreater(self.s.auto_click_interval, 0)

    def testCanLoadEnemiesFromFile(self):
        """testCanLoadEnemiesFromFile: should be able to load enemies from a file"""
        loader = model.EnemyLoader(ENEMY_FILE)
        enemies = loader.enemies
        self.assertGreater(len(enemies), 0)
        e = enemies[0]
        self.assertNotEquals(e.class_name, '')
        self.assertNotEquals(e.low_name, '')
        self.assertNotEquals(e.medium_name, '')
        self.assertNotEquals(e.high_name, '')
        self.assertGreater(e.value, 0)
        self.assertGreater(e.hp, 0)
        self.assertGreater(e.click_damage, 0)
        self.assertGreater(e.critical_click_multiplier, 0)
        self.assertGreater(e.critical_click_probability, 0)
        self.assertNotEquals(e.damage_per_second, None)
        self.assertNotEquals(e.auto_click_interval, None)
        self.assertNotEquals(e.hit_name, None)
        self.assertNotEquals(e.hit_per_second_name, None)
        self.assertEqual(len(e.wave_multipliers), 11)

    def testCanLoadPowerUps(self):
        """testCanLoadPowerUps: should be able to load power ups"""
        loader = model.PowerUpLoader(POWER_UP_FILE)
        power_ups = loader.power_ups
        self.assertGreater(len(power_ups), 0)
        p = power_ups[0]
        self.assertNotEqual(p.name, '')
        self.assertNotEqual(p.class_name, '')
        self.assertNotEqual(p.description, '')
        self.assertGreater(p.cost, 0)
        self.assertIsInstance(p.click_amount, float)
        self.assertIsInstance(p.aof_increase, float)
        self.assertIsInstance(p.hps_increase, float)
        self.assertIsInstance(p.hp_increase, float)
        self.assertIsInstance(p.ac_increase, float)


class TestCombat(unittest.TestCase):
    """Test of the Combat class"""

    def setUp(self):
        """Set up the tests"""
        self.p = model.Player()
        self.p.critical_click_probability = 0.0
        self.el = model.EnemyLoader(ENEMY_FILE)
        self.e = self.el.getEnemy(0)
        self.e.critical_click_probability = 0.0

    def tearDown(self):
        """Tear down the tests"""

    def testEnemiesCanBeReplicated(self):
        """testEnemiesCanBeReplicated: can replicate enemies"""
        e1 = self.el.getEnemy(0)
        e2 = self.el.getEnemy(0)
        self.assertEqual(e1.hp, e2.hp)
        #
        e1.hp += 1
        self.assertNotEqual(e1.hp, e2.hp)

    def testClickingDamages(self):
        """testClickingDamages: clicking should damage an enemy"""
        initial_hp = self.e.hp
        result = self.p.clickOn(self.e)
        self.assertEqual(result.actor, self.p)
        self.assertEqual(self.e.hp, initial_hp - self.p.click_damage)
        self.assertEqual(result.value, 0)
        self.assertEqual(result.damage, self.p.click_damage)
        self.assertEqual(result.damage_type, model.D_PLAYER_CLICK)

    def testDoingEnoughDamageKills(self):
        """testDoingEnoughDamageKills: doing enough damage kills enemy"""
        initial_hp = self.e.hp
        self.assertTrue(self.e.is_alive)
        self.p.click_damage = initial_hp * 2
        self.p.clickOn(self.e)
        self.assertEqual(self.e.hp, 0)
        self.assertFalse(self.e.is_alive)

    def testKillingGivesGold(self):
        """testKillingGivesGold: killing should yield gold"""
        initial_hp = self.e.hp
        self.p.click_damage = initial_hp * 2
        result = self.p.clickOn(self.e)
        self.assertNotEqual(self.e.value, 0)
        self.assertEqual(result.value, self.e.value)

    def testEnemyClickDamagePlayer(self):
        """testEnemyClickDamagePlayer: enemy can click damage the player"""
        initial_hp = self.p.hp
        self.p.armour_class = 1
        result = self.e.clickOn(self.p)
        self.assertEqual(self.p.hp, initial_hp - self.e.click_damage)
        self.assertEqual(result.damage_type, model.D_ENEMY_CLICK)

    def testPlayerACReducesDamage(self):
        """testPlayerACReducesDamage: adding to the AC of the player reduces damage"""
        initial_hp = self.p.hp
        self.p.ac = 2
        self.e.clickOn(self.p)
        self.assertEqual(self.p.hp, initial_hp - (self.e.click_damage - 1))

    def testHighEnoughACDoesNoDamage(self):
        """testHighEnoughACDoesNoDamage: if AC is too high then no damage occurs"""
        initial_hp = self.p.hp
        attack_damage = self.e.click_damage
        self.p.ac = attack_damage + 1
        self.e.clickOn(self.p)
        self.assertEqual(self.p.hp, initial_hp)

    def testSideKickClickDamageEnemy(self):
        """testSideKickClickDamageEnemy: a sidekick can click damage an enemy"""
        s = model.Sidekick()
        s.click_damage = 2
        initial_hp = self.e.hp
        result = s.clickOn(self.e)
        self.assertEqual(self.e.hp, initial_hp - s.click_damage)
        self.assertEqual(result.value, 0)

    def testSideKickGeneratesGold(self):
        """testSideKickGeneratesGold: sidekick will give gold"""
        s = model.Sidekick()
        s.click_damage = 2
        initial_hp = self.e.hp
        s.click_damage = initial_hp
        result = s.clickOn(self.e)
        self.assertEqual(self.e.hp, initial_hp - s.click_damage)
        self.assertEqual(result.value, self.e.value)
        self.assertEqual(result.damage_type, model.D_SIDE_CLICK)

    def testCriticalClickIncreases(self):
        """testCriticalClickIncreases: critical click increases damage"""
        initial_hp = self.e.hp
        self.p.critical_click_probability = 1.0
        result = self.p.clickOn(self.e)
        self.assertEqual(self.e.hp, initial_hp - self.p.click_damage * self.p.critical_click_multiplier)
        self.assertEqual(result.damage, self.p.click_damage * self.p.critical_click_multiplier)
        self.assertEqual(result.damage_type, model.D_PLAYER_CRITICAL_CLICK)
        initial_hp = self.e.hp
        self.p.critical_click_probability = 0.0
        self.p.clickOn(self.e)
        self.assertEqual(self.e.hp, initial_hp - self.p.click_damage)

    def testEnemiesCanCritical(self):
        """testEnemiesCanCritical: enemies can critical the player"""
        initial_hp = self.p.hp
        self.p.armour_class = 1
        self.e.critical_click_probability = 1.0
        result = self.e.clickOn(self.p)
        self.assertEqual(self.p.hp, initial_hp - self.e.click_damage * self.e.critical_click_multiplier)
        self.assertEqual(result.action, 'Critical Bites')
        self.assertEqual(result.damage_type, model.D_ENEMY_CRITICAL_CLICK)


class TestPowerUps(unittest.TestCase):
    """Test of the PowerUps class"""

    def setUp(self):
        """Set up the tests"""
        self.p = model.Player()
        self.p.critical_click_probability = 0.0
        self.el = model.EnemyLoader(ENEMY_FILE)
        self.e = self.el.getEnemy(0)
        self.e.critical_click_probability = 0.0
        #
        self.pl = model.PowerUpLoader(POWER_UP_FILE)

    def tearDown(self):
        """Tear down the tests"""

    def testPowerUpClick(self):
        """testPowerUpClick: can power up click"""
        pu = model.PowerUp()
        pu.click_amount = 1
        initial = self.p.click_damage
        self.p.addPowerUp(pu)
        self.assertEqual(self.p.click_damage, initial + 1)

    def testCanPowerUpAC(self):
        """testCanPowerUpAC: power up AC"""
        pu = model.PowerUp()
        pu.ac_increase = 1
        initial = self.p.ac
        self.p.addPowerUp(pu)
        self.assertEqual(self.p.ac, initial + 1)

    def testCanPowerUpHP(self):
        """testCanPowerUpHP: can power up HP"""
        pu = model.PowerUp()
        pu.hp_increase = 1
        initial = self.p.hp
        self.p.addPowerUp(pu)
        self.assertEqual(self.p.hp, initial + 1)

    def testPowerUpHPWhenLowerHealth(self):
        """testPowerUpHPWhenLowerHealth: when adding HP with low health should work OK"""
        pu = model.PowerUp()
        pu.hp_increase = 10
        self.p.hp -= 5
        initial = self.p.max_hp
        self.p.addPowerUp(pu)
        self.assertEqual(self.p.max_hp, initial + 10)
        self.assertEqual(self.p.hp, self.p.max_hp)

    def testCanPowerUpAOF(self):
        """testCanPowerUpAOF: can power up AOF"""
        pu = model.PowerUp()
        pu.aof_increase = 1
        initial = self.p.click_aof_damage
        self.p.addPowerUp(pu)
        self.assertEqual(self.p.click_aof_damage, initial + 1)

    def testCanPowerUpHPS(self):
        """testCanPowerUpHPS: can power up HPS"""
        pu = model.PowerUp()
        pu.hps_increase = 1
        initial = self.p.heal_per_second
        self.p.addPowerUp(pu)
        self.assertEqual(self.p.heal_per_second, initial + 1)

    def testMultiplePowerUps(self):
        """testMultiplePowerUps: can have multiple power ups"""
        initial_click = self.p.click_damage
        initial_hp = self.p.hp
        initial_ac = self.p.ac
        #
        pu = model.PowerUp()
        pu.click_amount = 1
        self.p.addPowerUp(pu)
        #
        pu = model.PowerUp()
        pu.hp_increase = 1
        self.p.addPowerUp(pu)
        #
        pu = model.PowerUp()
        pu.ac_increase = 1
        self.p.addPowerUp(pu)
        #
        self.assertEqual(self.p.ac, initial_ac + 1)
        self.assertEqual(self.p.hp, initial_hp + 1)
        self.assertEqual(self.p.click_damage, initial_click + 1)

    def testSideKickPowerUp(self):
        """testSideKickPowerUp: side kicks can power up"""
        s = model.Sidekick()
        s.click_damage = 2
        pu = model.PowerUp()
        pu.click_amount = 1
        initial_hp = self.e.hp
        #
        s.addPowerUp(pu)
        result = s.clickOn(self.e)
        self.assertEqual(self.e.hp, initial_hp - 3)
        self.assertEqual(result.value, 0)


class TestTimeBasedCombat(unittest.TestCase):
    """Test of the TimeBasedCombat class"""

    def setUp(self):
        """Set up the tests"""
        self.p = model.Player()
        self.p.critical_click_probability = 0.0
        self.el = model.EnemyLoader(ENEMY_FILE)
        self.e = self.el.getEnemy(0)
        self.e.critical_click_probability = 0.0

    def tearDown(self):
        """Tear down the tests"""

    def testAttackOnlyAliveEnemies(self):
        """testAttackOnlyAliveEnemies: only attack alive enemies"""
        e1, e2 = self.el.getEnemy(0), self.el.getEnemy(0)
        e1.is_alive = False
        e2.hp = 200
        self.p.auto_click_interval = 1.0
        for i in range(100):
            result = self.p.tickClock(1.0, [e1, e2])
            self.assertEqual(result[0].target, e2)

    def testCanTickWithAllDead(self):
        """testCanTickWithAllDead: shold be able to tick with all dead"""
        e1, e2 = self.el.getEnemy(0), self.el.getEnemy(0)
        e1.is_alive = False
        e2.is_alive = False
        self.p.auto_click_interval = 1.0
        for i in range(100):
            result = self.p.tickClock(1.0, [e1, e2])
            self.assertEqual(result[0].action, '')

    def testTickWithoutAction(self):
        """testTickWithoutAction: can tick without action"""
        initial_hp = self.p.hp
        damage = self.e.click_damage
        self.e.auto_click_interval = 1.0
        results = self.e.tickClock(0.9, [self.p])
        self.assertEqual(results[0].target, None)
        self.assertEqual(results[0].action, '')
        self.assertEqual(results[0].damage, 0)
        self.assertEqual(len(results), 1)

    def testTickWhenShouldAttack(self):
        """testTickWhenShouldAttack: should attack after a while"""
        initial_hp = self.p.hp
        damage = self.e.click_damage
        self.e.auto_click_interval = 1.0
        results = self.e.tickClock(1.0, [self.p])
        self.assertEqual(results[0].target, self.p)
        self.assertEqual(results[0].action, self.e.hit_name)
        self.assertEqual(results[0].damage, damage)
        self.assertEqual(self.p.hp, initial_hp - damage)
        self.assertEqual(len(results), 1)

    def testMultipleTicksOverTime(self):
        """testMultipleTicksOverTime: should attack multiple times"""
        self.p.hp = 100.0
        initial_hp = self.p.hp
        damage = self.e.click_damage
        self.e.auto_click_interval = 1.0
        #
        for i in range(101):
            self.e.tickClock(0.1, [self.p])
        #
        self.assertEqual(self.p.hp, initial_hp - 10.0 * damage)

    def testSickKickNoAttack(self):
        """testSickKickNoAttack: sidekick needs time to attack over time"""
        s = model.Sidekick()
        initial_hp = self.e.hp
        damage = s.click_damage = 2.0
        s.auto_click_interval = 1.0
        results = s.tickClock(0.9, [self.e])
        self.assertEqual(results[0].target, None)
        self.assertEqual(results[0].action, '')
        self.assertEqual(results[0].damage, 0)
        self.assertEqual(len(results), 1)

    def testSickKickAttack(self):
        """testSickKickAttack: sidekick needs attack over time"""
        s = model.Sidekick()
        initial_hp = self.e.hp
        damage = s.click_damage = 2.0
        s.auto_click_interval = 1.0
        results = s.tickClock(1.0, [self.e])
        self.assertEqual(results[0].target, self.e)
        self.assertEqual(results[0].action, 'Hit')
        self.assertEqual(results[0].damage, 2.0)
        self.assertEqual(len(results), 1)

    def testSideKickAttackRandom(self):
        """testSideKickAttackRandom: sidekick can attack random"""
        enemies = [self.el.getEnemy(0), self.el.getEnemy(0)]
        s = model.Sidekick()
        initial_hp = self.e.hp
        damage = s.click_damage = 2.0
        s.auto_click_interval = 1.0
        results = s.tickClock(1.0, enemies)
        self.assertIn(results[0].target, enemies)
        self.assertEqual(results[0].action, 'Hit')
        self.assertEqual(results[0].damage, 2.0)
        self.assertEqual(len(results), 1)

    def testSideKickRandomNotTheSame(self):
        """testSideKickRandomNotTheSame: should pick a truly random number"""
        enemies = [self.el.getEnemy(0), self.el.getEnemy(0)]
        s = model.Sidekick()
        initial_hp = self.e.hp
        damage = s.click_damage = 0.002
        s.auto_click_interval = 1.0
        #
        min_value = 100
        max_value = -100
        #
        for i in range(100):
            results = s.tickClock(1.0, enemies)
            posn = enemies.index(results[0].target)
            min_value = min(min_value, posn)
            max_value = max(max_value, posn)
        #
        self.assertEqual(min_value, 0)
        self.assertEqual(max_value, 1)

    def testSideKickAOFDamage(self):
        """testSideKickAOFDamage: side kick can do AOF damage"""
        enemies = [self.el.getEnemy(0), self.el.getEnemy(0)]
        s = model.Sidekick()
        initial_hp = self.e.hp
        damage = s.click_damage = 2.0
        s.click_aof_damage = 1.0
        s.auto_click_interval = 1.0
        results = s.tickClock(1.0, enemies)
        self.assertEqual(len(results), 2)
        for result in results:
            self.assertIn(result.target, enemies)
            self.assertEqual(result.action, 'Hit')
            self.assertEqual(result.damage, 2.0)
            self.assertEqual(result.damage_type, model.D_SIDE_AOF)

    def testSideKickAOFExcludesPlayer(self):
        """testSideKickAOFExcludesPlayer: AOF should not affect the player"""
        enemies = [self.el.getEnemy(0), self.el.getEnemy(0), self.p]
        player_hp = self.p.hp
        s = model.Sidekick()
        initial_hp = self.e.hp
        damage = s.click_damage = 2.0
        s.click_aof_damage = 1.0
        s.auto_click_interval = 1.0
        results = s.tickClock(1.0, enemies)
        self.assertEqual(len(results), 2)
        for result in results:
            self.assertIn(result.target, enemies)
            self.assertEqual(result.action, 'Hit')
            self.assertEqual(result.damage, 2.0)
        #
        self.assertEqual(self.p.hp, player_hp)


class TestDamageOverTime(unittest.TestCase):
    """Test of the DamageOverTime class"""

    def setUp(self):
        """Set up the tests"""
        self.p = model.Player()
        self.p.critical_click_probability = 0.0
        self.el = model.EnemyLoader(ENEMY_FILE)
        self.e = self.el.getEnemy(0)
        self.e.critical_click_probability = 0.0

    def tearDown(self):
        """Tear down the tests"""

    def testEnemyCanDoDOT(self):
        """testEnemyCanDoDOT: do damage over time"""
        initial_hp = self.p.hp
        self.e.damage_per_second = 2.0
        self.e.click_damage = 1.0
        result = self.e.clickOn(self.p)
        self.assertEqual(result.action, self.e.hit_per_second_name)
        self.assertEqual(result.damage, self.e.click_damage)
        self.assertEqual(self.p.hp, initial_hp - self.e.click_damage)

        #
        # First half second nothing happens
        results = self.p.tickClock(self.p.damage_over_time_interval / 2.0, [self.e])
        self.assertEqual(results[0].action, '')
        self.assertEqual(self.p.hp, initial_hp - self.e.click_damage)

        #
        # Now something should happen
        results = self.p.tickClock(self.p.damage_over_time_interval / 2.0, [self.e])
        self.assertEqual(len(results), 2)
        self.assertNotEqual(results[1].action, '')
        self.assertEqual(results[1].target, self.p)
        self.assertEqual(results[1].damage_type, model.D_ENEMY_DOT)
        self.assertEqual(self.p.hp, initial_hp - self.e.click_damage - self.e.damage_per_second)

    def testDOTContiues(self):
        """testDOTContiues: DOT continues over time"""
        initial_hp = self.p.hp = 100.0
        self.e.damage_per_second = 2.0
        self.e.click_damage = 1.0
        result = self.e.clickOn(self.p)
        #
        for i in range(10):
            self.p.tickClock(self.p.damage_over_time_interval, [self.e])
        #
        self.assertEqual(self.p.hp, initial_hp - self.e.click_damage - 10 * self.e.damage_per_second)

    def testCanCureDamageAndDOT(self):
        """testCanCureDamageAndDOT: should be able to cure damage and DOT"""
        initial_hp = self.p.hp = 100.0
        self.e.damage_per_second = 2.0
        self.e.click_damage = 1.0
        result = self.e.clickOn(self.p)
        #
        self.p.cureAll()
        self.assertEqual(self.p.hp, self.p.max_hp)
        self.p.tickClock(1000, [self.e])
        self.assertEqual(self.p.hp, self.p.max_hp)

    def testDOTStacks(self):
        """testDOTStacks: DOT stacks"""
        initial_hp = self.p.hp = 100.0
        self.e.damage_per_second = 2.0
        self.e.click_damage = 1.0
        result = self.e.clickOn(self.p)
        result = self.e.clickOn(self.p)
        self.p.tickClock(self.p.damage_over_time_interval, [self.e])
        self.assertEqual(self.p.hp, initial_hp - 2 * self.e.click_damage - 2 * self.e.damage_per_second)


class TestHealing(unittest.TestCase):
    """Test of the Healing class"""

    def setUp(self):
        """Set up the tests"""
        self.p = model.Player()
        self.p.critical_click_probability = 0.0
        self.el = model.EnemyLoader(ENEMY_FILE)
        self.e = self.el.getEnemy(0)
        self.e.critical_click_probability = 0.0

    def tearDown(self):
        """Tear down the tests"""

    def testSideKickCanHeal(self):
        """testSideKickCanHeal: should be able to heal the player"""
        initial_hp = self.p.hp = 100.0
        self.p.hp -= 10
        s = model.Sidekick()
        s.heal_per_second = 1.0
        #
        result = s.clickOn(self.p)
        self.assertEqual(self.p.hp, initial_hp - 10 + 1)
        self.assertEqual(result.target, self.p)
        self.assertEqual(result.action, 'Heal')
        self.assertEqual(result.damage_type, model.D_SIDE_HEAL)
        self.assertEqual(result.damage, 1.0)

    def testSideKickHealOriginalHP(self):
        """testSideKickHealOriginalHP: sidekick cannot heal byond original HP"""
        initial_hp = self.p.hp = 100.0
        self.p.hp -= 10
        s = model.Sidekick()
        s.heal_per_second = 1.0
        #
        for i in range(100):
            s.clickOn(self.p)
        self.assertEqual(self.p.hp, initial_hp)

    def testSideKickHealOverTime(self):
        """testSideKickHealOverTime: sidekick can heal over time"""
        initial_hp = self.p.hp = 100.0
        self.p.hp -= 10
        s = model.Sidekick()
        s.heal_per_second = 1.0
        s.auto_click_interval = 2.0
        #
        results = s.tickClock(s.auto_click_interval / 2.0, [self.p])
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0].action, '')
        #
        results = s.tickClock(s.auto_click_interval / 2.0, [self.p])
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0].action, 'Heal')
        self.assertEqual(results[0].target, self.p)
        self.assertEqual(results[0].damage, 1.0)

    def testSideKickHealExcludesEnemies(self):
        """testSideKickHealExcludesEnemies: sidekick healing should just be for player"""
        initial_hp = self.p.hp = 100.0
        initial_enemy_hp = self.e.hp = 100.0
        self.p.hp -= 10
        self.e.hp -= 10
        s = model.Sidekick()
        s.heal_per_second = 1.0
        s.auto_click_interval = 2.0
        #
        results = s.tickClock(s.auto_click_interval, [self.p, self.e])
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0].action, 'Heal')
        self.assertEqual(results[0].target, self.p)
        self.assertEqual(results[0].damage, 1.0)
        #
        self.assertEqual(self.p.hp, initial_hp - 10 + 1)
        self.assertEqual(self.e.hp, initial_hp - 10)


class TestBattle(unittest.TestCase):
    """Test of the Battle class"""
    
    def setUp(self):
        """Set up the tests"""
        self.p = model.Player()
        self.p.critical_click_probability = 0.0
        self.el = model.EnemyLoader(ENEMY_FILE)

    def tearDown(self):
        """Tear down the tests"""

    def testGetEnemiesInWave(self):
        """testGetEnemiesInWave: can get enemies in wave"""
        enemies = self.el.getEnemiesInWave(1)
        self.assertEqual(len(enemies), 2)
        enemies = self.el.getEnemiesInWave(2)
        self.assertEqual(len(enemies), 4)
        enemies = self.el.getEnemiesInWave(3)
        self.assertEqual(len(enemies), 6)

    def testChooseANumberOfEnemies(self):
        """testChooseANumberOfEnemies: can choose a number of enemies"""
        for wave in range(1, 11):
            enemies = self.el.getEnemyGroup(wave, 10)
            enemies_in_wave = [e.class_name for e in self.el.getEnemiesInWave(wave)]
            #
            self.assertEqual(len(enemies), 10)
            for enemy in enemies:
                self.assertIn(enemy.class_name, enemies_in_wave)

    def testGetEnemiesInWavesIncreaseStrength(self):
        """testGetEnemiesInWavesIncreaseStrength: enemies increase strength with waves"""
        e1 = self.el.getEnemiesInWave(1)
        e2 = self.el.getEnemiesInWave(2)
        e3 = self.el.getEnemiesInWave(3)
        for e_in_1 in e1:
            for e_in_2 in e2:
                if e_in_1.class_name == e_in_2.class_name:
                    self.assertGreater(e_in_2.hp, e_in_1.hp)
                    self.assertEqual(e_in_1.real_name, '%s %s' % (e_in_1.low_name, e_in_1.class_name))
                    self.assertEqual(e_in_2.real_name, '%s %s' % (e_in_1.medium_name, e_in_1.class_name))
                    for e_in_3 in e3:
                        if e_in_1.class_name == e_in_3.class_name:
                            self.assertGreater(e_in_2.hp, e_in_1.hp)
                            self.assertEqual(e_in_3.real_name, '%s %s' % (e_in_1.high_name, e_in_1.class_name))

    def testCheckEnemiesAlive(self):
        """testCheckEnemiesAlive: check a group for enemies being alive"""
        g = self.el.getEnemiesInWave(1)
        self.assertTrue(g.hasSomeAlive())
        for i in g:
            i.is_alive = False
        self.assertFalse(g.hasSomeAlive())

    def testCanTickEnemiesInGroup(self):
        """testCanTickEnemiesInGroup: can tick enemies in group"""
        enemies = self.el.getEnemyGroup(1, 2)
        for e in enemies:
            e.auto_click_interval = 1.0
        results = enemies.tickClock(1.0, [self.p])
        #
        self.assertEqual(len(results), 2)
        for result in results:
            self.assertEqual(result.target, self.p)

    def testDoNotTickDeadEnemies(self):
        """testDoNotTickDeadEnemies: dead enemies should not tick"""
        enemies = self.el.getEnemyGroup(1, 2)
        for e in enemies:
            e.auto_click_interval = 1.0
            e.is_alive = False
        results = enemies.tickClock(1.0, [self.p])
        #
        self.assertEqual(len(results), 0)


class TestPurse(unittest.TestCase):
    """Test of the Purse class"""

    def setUp(self):
        """Set up the tests"""
        self.p = model.Player()

    def tearDown(self):
        """Tear down the tests"""

    def testCanAddMoney(self):
        """testCanAddMoney: can add money"""
        initial = self.p.purse
        self.p.addMoney(10)
        self.assertEqual(self.p.purse, initial + 10)

    def testCanSpendMoney(self):
        """testCanSpendMoney: can spend money"""
        initial = self.p.purse = 20
        self.p.addMoney(-10)
        self.assertEqual(self.p.purse, initial - 10)

    def testMoneyCannotIncrease(self):
        """testMoneyCannotIncrease: cannot increase money over max"""
        initial = self.p.purse
        max_purse = self.p.max_purse_size
        self.p.addMoney(10000)
        self.assertEqual(self.p.purse, max_purse)
        self.p.addMoney(-2 * max_purse)
        self.assertEqual(self.p.purse, 0)


class TestPowerUpPurchase(unittest.TestCase):
    """Test of the PowerUpPurchase class"""

    def setUp(self):
        """Set up the tests"""
        self.p = model.Player()
        self.pl = model.PowerUpLoader(POWER_UP_FILE)

    def tearDown(self):
        """Tear down the tests"""

    def testCanGetClassNames(self):
        """testCanGetClassNames: should be able to get class names"""
        names = self.pl.classes.keys()
        self.assertGreater(len(names), 0)

    def testGetItemsInClass(self):
        """testGetItemsInClass: can get items in class"""
        names = self.pl.classes.keys()
        values = self.pl.classes[names[0]]
        self.assertGreater(len(values), 0)

    def testFirstItemIsAvailable(self):
        """testFirstItemIsAvailable: first item in class should be available"""
        names = self.pl.classes.keys()
        values = self.pl.classes[names[0]]
        self.assertGreater(len(values), 1)
        self.assertTrue(values[0].is_available)
        for item in values[1:]:
            self.assertFalse(item.is_available)

    def testAfterBuyingItemItNotAvailableAndNextIs(self):
        """testAfterBuyingItemItNotAvailableAndNextIs: normal buying workflow"""
        names = self.pl.classes.keys()
        values = self.pl.classes[names[0]]
        self.assertGreater(len(values), 1)
        self.assertTrue(values[0].is_available)
        #
        values[0].buyItem()
        self.assertFalse(values[0].is_available)
        self.assertTrue(values[1].is_available)
        #
        for item in values[2:]:
            self.assertFalse(item.is_available)

    def testBuyLastItem(self):
        """testBuyLastItem: when buy last item then non are available"""
        names = self.pl.classes.keys()
        values = self.pl.classes[names[0]]
        #
        self.assertTrue(values.hasItemsAvailable())
        for item in values:
            item.buyItem()
        self.assertFalse(values.hasItemsAvailable())

    def testGetFirstAvailable(self):
        """testGetFirstAvailable: get the first available item"""
        names = self.pl.classes.keys()
        values = self.pl.classes[names[0]]
        #
        item = values.getFirstAvailable()
        self.assertEqual(values.index(item), 0)
        item.buyItem()
        new_item = values.getFirstAvailable()
        self.assertEqual(values.index(new_item), 1)

    def testGetAvailableItems(self):
        """testGetAvailableItems: get the items available with this cash"""
        items = self.pl.getItemsAvailable(0)
        self.assertEqual(len(items), 0)
        items = self.pl.getItemsAvailable(1000000)
        self.assertEqual(len(items), len(self.pl.classes))

    def testPlayerBuyingPowerUp(self):
        """testPlayerBuyingPowerUp: player can buy power ups"""
        self.p.purse = self.p.max_purse_size = 1000
        items = self.pl.getItemsAvailable(self.p.purse)
        item = items[0]
        self.p.addPowerUp(item)
        #
        self.assertFalse(item.is_available)
        self.assertEqual(self.p.purse, self.p.max_purse_size - item.cost)

    def testCanResetPowerUps(self):
        """testCanResetPowerUps: can reset power ups"""
        items = self.pl.getItemsAvailable(1000)
        original = len(items)
        self.assertNotEqual(original, 0)
        #
        while True:
            items = self.pl.getItemsAvailable(1000)
            if not items:
                break
            for item in items:
                item.buyItem()
        #
        self.pl.resetPowerUps()
        new_items = self.pl.getItemsAvailable(1000)
        self.assertEqual(original, len(new_items))

    def testCanTrackTotalSpend(self):
        """testCanTrackTotalSpend: can track total spend"""
        items = self.pl.getItemsAvailable(1000)
        total_cost = 0
        for item in items[:5]:
            self.p.addPowerUp(item)
            total_cost += item.cost
        self.assertEqual(self.p.total_spent, total_cost)
        self.assertEqual(self.p.total_power_ups, 4)



class TestSideKickActivation(unittest.TestCase):
    """Test of the SideKickActivation class"""

    def setUp(self):
        """Set up the tests"""
        self.p = model.Player()
        self.p.critical_click_probability = 0.0
        self.el = model.EnemyLoader(ENEMY_FILE)
        self.e = self.el.getEnemy(0)
        self.pl = model.PowerUpLoader(POWER_UP_FILE)
        self.s1 = self.p.sidekicks['Sidekick-A']
        self.s2 = self.p.sidekicks['Sidekick-B']
        self.s3 = self.p.sidekicks['Sidekick-C']
        self.s1.auto_click_interval = 1.0
        self.s2.auto_click_interval = 1.0
        self.s3.auto_click_interval = 1.0

    def tearDown(self):
        """Tear down the tests"""

    def testSideKicksInactive(self):
        """testSideKicksInactive: sidekicks are inactive"""
        self.assertFalse(self.s1.is_alive)
        self.assertFalse(self.s2.is_alive)
        self.assertFalse(self.s3.is_alive)

    def testAddPowerUpActivates(self):
        """testAddPowerUpActivates: adding a power up activates the sidekick"""
        s1 = self.pl.classes['Sidekick-A'].getFirstAvailable()
        self.p.addPowerUp(s1)
        self.assertTrue(self.s1.is_alive)
        self.assertFalse(self.s2.is_alive)
        self.assertFalse(self.s3.is_alive)
        #
        s2 = self.pl.classes['Sidekick-B'].getFirstAvailable()
        self.p.addPowerUp(s2)
        self.assertTrue(self.s2.is_alive)
        self.assertFalse(self.s3.is_alive)
        #
        s3 = self.pl.classes['Sidekick-C'].getFirstAvailable()
        self.p.addPowerUp(s3)
        self.assertTrue(self.s3.is_alive)

    def testTickingTicksActive(self):
        """testTickingTicksActive: ticking should tick active sidekicks"""
        self.p.auto_click_interval = 0.0
        results = self.p.tickClock(1, [self.e])
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0].actor, self.p)
        self.assertEqual(results[0].action, '')
        #
        self.s1.is_alive = True
        results = self.p.tickClock(1, [self.e])
        self.assertEqual(len(results), 2)
        self.assertEqual(results[0].actor, self.p)
        self.assertEqual(results[0].action, '')
        self.assertEqual(results[1].actor, self.s1)
        self.assertEqual(results[1].target, self.e)
        #
        self.s2.is_alive = True
        self.s3.is_alive = True
        results = self.p.tickClock(1, [self.e])
        self.assertEqual(len(results), 4)
        self.assertEqual(results[0].actor, self.p)
        self.assertEqual(results[0].action, '')
        self.assertEqual(results[1].actor, self.s1)
        self.assertEqual(results[1].target, self.e)
        #
        self.assertEqual(results[2].actor, self.s2)
        self.assertEqual(results[2].target, self.e)
        #
        # Sidekick C is the healer
        self.assertEqual(results[3].actor, self.s3)
        self.assertEqual(results[3].target, self.e)

    def testAddSKPowerUpAffectsThem(self):
        """testAddSKPowerUpAffectsThem: adding  power up affects the right sidekick"""
        self.assertFalse(self.s2.is_alive)
        pu = self.pl.classes['Sidekick-B'].getFirstAvailable()
        pu.buyItem()
        self.p.addPowerUp(pu)
        #
        self.assertTrue(self.s2.is_alive)
        pu2 = self.pl.classes['Sidekick-B'].getFirstAvailable()
        self.assertNotEqual(pu, pu2)
        #
        initial_player = self.p.click_damage
        initial = self.s2.click_damage
        increase = pu2.click_amount
        self.assertNotEqual(increase, 0)
        self.p.addPowerUp(pu2)
        #
        self.assertEqual(self.p.click_damage, initial_player)
        self.assertEqual(self.s2.click_damage, initial + increase)

    def testTickingCanFindPlayer(self):
        """testTickingCanFindPlayer: should find player"""
        self.p.sidekicks['Sidekick-C'].is_alive = True
        self.p.sidekicks['Sidekick-C'].auto_click_interval = 1.0
        self.p.sidekicks['Sidekick-C'].heal_per_second = 10.0
        self.p.hp = 0.0
        results = self.p.tickClock(1.0, [self.e])
        #
        # Should have healed
        self.assertEqual(self.p.hp, 10.0)



if __name__ == '__main__':
    unittest.main()